import unittest

from mintersdk.sdk.check import MinterCheck
from mintersdk.sdk.transactions import MinterTx


class TestMinterCheck(unittest.TestCase):

    def setUp(self):
        self.PRIVATE_KEY = '64e27afaab363f21eec05291084367f6f1297a7b280d69d672febecda94a09ea'
        self.ADDRESS = 'Mxa7bc33954f1ce855ed1a8c768fdd32ed927def47'
        self.PASSPHRASE = 'pass'
        self.VALID_CHECK = 'Mcf8ae8334383002830f423f8a4d4e5400000000000000888ac7230489e800008a4d4e5400000000000000b841497c5f3e6fc182fd1a791522a9ef7576710bdfbc86fdbf165476ef220e89f9ff1380f93f2d9a2f92fdab0edc1e2605cc2c69b707cd404b2cb1522b7aba4defd5001ba083c9945169f0a7bbe596973b32dc887608780580b1d3bc7b188bedb3bd385594a047b2d5345946ed5498f5bee713f86276aac046a5fef820beaee77a9b6f9bc1df'
        self.VALID_PROOF = 'da021d4f84728e0d3d312a18ec84c21768e0caa12a53cb0a1452771f72b0d1a91770ae139fd6c23bcf8cec50f5f2e733eabb8482cf29ee540e56c6639aac469600'
        self.CHECK = MinterCheck(
            nonce=480,
            due_block=999999,
            coin='MNT',
            value=10,
            passphrase=self.PASSPHRASE,
            chain_id=MinterTx.TESTNET_CHAIN_ID,
            gas_coin='MNT'
        )

    def test_check(self):
        check = self.CHECK.sign(self.PRIVATE_KEY)

        self.assertEqual(check, self.VALID_CHECK)

    def test_proof(self):
        proof = MinterCheck.proof(
            address=self.ADDRESS,
            passphrase=self.PASSPHRASE
        )

        self.assertEqual(proof, self.VALID_PROOF)

    def test_fromraw(self):
        check = MinterCheck.from_raw(rawcheck=self.VALID_CHECK)

        self.assertEqual(
            check.owner,
            'Mxce931863b9c94a526d94acd8090c1c5955a6eb4b'
        )
        self.assertEqual(check.gas_coin, self.CHECK.gas_coin)



"""
@author: Roman Matusevich
"""
import hashlib
import hmac

import sslcrypto
from mnemonic.mnemonic import Mnemonic
from mintersdk import MinterHelper, PREFIX_PUBKEY, PREFIX_ADDR


class MinterWallet(object):
    """
    Minter wallet class
    """

    # Amount of entropy bits (BIP44)
    entropy_bits = 128

    # Address path for creating wallet from the seed (BIP44)
    seed_address_path = "m/44'/60'/0'/0/0"

    # Master seed
    master_seed = b'Bitcoin seed'

    # Curve data
    curve = sslcrypto.ecc.get_curve('secp256k1')
    pub_key_len = curve._backend.public_key_length

    @classmethod
    def create(cls, mnemonic=None):
        """
        Create Minter wallet
        Args:
            mnemonic (str): Mnemonic phrase
        Returns:
            dict
        """

        # Create mnemonic phrase if None
        if not mnemonic:
            _mnemonic = Mnemonic(language='english')
            mnemonic = _mnemonic.generate(cls.entropy_bits)

        if len(mnemonic.split(' ')) != 12:
            raise Exception('Mnemonic phrase should have 12 words.')

        # Mnemonic to seed (bytes)
        seed = Mnemonic.to_seed(mnemonic, '')

        # Generate master key (key, hmac_key) from master seed
        _I = hmac.new(cls.master_seed, seed, hashlib.sha512).hexdigest()
        master_key = (int(_I[:64], 16), bytes.fromhex(_I[64:]))

        # Get child keys from master key by path
        keys = cls.from_path(
            root_key=master_key, path=cls.seed_address_path
        )

        # Get private key
        private_key = format(keys[-1][0], '0x')
        # Get public key from private
        public_key = cls.get_public_from_private(private_key)
        # Get address from public key
        address = cls.get_address_from_public_key(public_key)

        return {
            'address': address,
            'private_key': private_key,
            'mnemonic': mnemonic,
            'seed': seed.hex()
        }

    @classmethod
    def get_public_from_private(cls, private_key):
        """
        Get public key from private key
        Args:
            private_key (str): hex bytes of private key
        Returns:
            str
        """
        # Get public key from private
        public_key = cls.curve.private_to_public(
            int(private_key, 16).to_bytes(length=32, byteorder='big')
        )
        public_key = public_key.hex()[2:]

        return MinterHelper.prefix_add(public_key, PREFIX_PUBKEY)

    @classmethod
    def get_address_from_public_key(cls, public_key):
        """
        Args:
            public_key (str)
        Returns:
            str
        """
        # Create keccak hash
        _keccak = MinterHelper.keccak_hash(
            bytes.fromhex(MinterHelper.prefix_remove(public_key))
        )

        return MinterHelper.prefix_add(_keccak[-40:], PREFIX_ADDR)

    @staticmethod
    def parse_path(path):
        """
        Parsing seed address path.
        Method was ported from 'two1.bitcoin.crypto'
        Args:
            path (str): Seed address path
        Returns:
            list
        """
        if isinstance(path, str):
            # Remove trailing "/"
            p = path.rstrip("/").split("/")
        elif isinstance(path, bytes):
            p = path.decode('utf-8').rstrip("/").split("/")
        else:
            p = list(path)

        return p

    @classmethod
    def from_parent(cls, parent_key, index):
        """
        Generate child private key from parent private key.
        Method was ported from 'two1.bitcoin.crypto'.
        Method is suitable only for private keys. To use full functionality,
        you should install 'two1' package.
        Args:
            parent_key (tuple(int, bytes)): Tuple of key and hmac_key
            index (int): Child index
        Returns:
            tuple(int, bytes): Child key
        """

        if index < 0 or index > 0xffffffff:
            raise ValueError("index is out of range: 0 <= index <= 2**32 - 1")

        # Get curve n parameter.
        curve_n = int(cls.curve.params['n'])

        # Unpack parent key
        parent_key, hmac_key = parent_key

        if index & 0x80000000:
            hmac_data = b'\x00' + parent_key.to_bytes(length=32, byteorder='big')
        else:
            # Create default curve public key from private
            public_key = cls.curve.private_to_public(
                parent_key.to_bytes(length=32, byteorder='big')
            )

            # Get public key coordinates
            x, y = cls.curve.decode_public_key(public_key)
            x = int.from_bytes(x, byteorder='big')
            y = int.from_bytes(y, byteorder='big')

            # Generate hmac data
            hmac_data = (
                bytes([(y & 0x1) + 0x02]) +
                x.to_bytes(cls.pub_key_len, 'big')
            )
        hmac_data += index.to_bytes(length=4, byteorder='big')

        I = hmac.new(hmac_key, hmac_data, hashlib.sha512).digest()
        Il, Ir = I[:32], I[32:]

        parse_Il = int.from_bytes(Il, 'big')
        if parse_Il >= curve_n:
            return None

        child_key = (parse_Il + parent_key) % curve_n
        if child_key == 0:
            # Incredibly unlucky choice
            return None

        return child_key, Ir

    @classmethod
    def from_path(cls, root_key, path):
        """
        Generate keys from path.
        Method was ported from 'two1.bitcoin.crypto'
        Args:
            root_key (tuple(int, bytes)): Tuple of key and hmac_key
            path (str): Seed address path
        Returns:
            list(tuple(int, bytes)): List of tuples (key, hmac_key)
        """
        p = cls.parse_path(path)

        if p[0] == "m":
            p = p[1:]

        keys = [root_key]
        for i in p:
            if isinstance(i, str):
                hardened = i[-1] == "'"
                index = int(i[:-1], 0) | 0x80000000 if hardened else int(i, 0)
            else:
                index = i
            k = keys[-1]

            keys.append(cls.from_parent(parent_key=k, index=index))

        return keys



import os
import random
import decimal
import string
import hashlib
import sha3

import pyqrcode
from deprecated import deprecated


# Number of PIP in 1 BIP
PIP = 1000000000000000000

# Prefixes
PREFIX_ADDR = 'Mx'
PREFIX_PUBKEY = 'Mp'
PREFIX_CHECK = 'Mc'
PREFIX_TX = 'Mt'


@deprecated("Use 'to_bip', 'to_pip' shortcuts or MinterHelper methods")
class MinterConvertor:
    """
    Class contains different converters
    """

    # PIP in BIP
    DEFAULT = 1000000000000000000

    @classmethod
    def convert_value(cls, value, to, prec=33):
        """
        Convert values from/to pip/bip.
        Args:
            value (string|int|Decimal|float): value to convert
            to (string): coin to convert value to
            prec (int): decimal context precision (decimal number length)
        Returns:
            int|Decimal
        """
        # Get default decimal context
        context = decimal.getcontext()
        # Set temporary decimal context for calculation
        decimal.setcontext(
            decimal.Context(prec=prec, rounding=decimal.ROUND_DOWN)
        )

        # PIP in BIP in Decimal
        default = decimal.Decimal(str(cls.DEFAULT))
        # Value in Decimal
        value = decimal.Decimal(str(value))

        # Make conversion
        if to == 'pip':
            value = int(value * default)
        elif to == 'bip':
            value /= default

        # Reset decimal context to default
        decimal.setcontext(context)

        return value

    @classmethod
    def encode_coin_name(cls, symbol):
        """
        Add nulls to coin name
        Args:
            symbol (string): coin symbol
        Returns:
            string
        """
        return symbol + chr(0) * (10 - len(symbol))

    @classmethod
    def decode_coin_name(cls, symbol):
        """
        Args:
            symbol (bytes|str)
        Returns:
            string
        """

        if hasattr(symbol, 'decode'):
            symbol = symbol.decode()

        return symbol.replace(chr(0), '')


class MinterHelper:
    """
    Class which contains different helpers
    """

    @staticmethod
    def keccak_hash(data, digest_bits=256):
        """
        Create Keccak hash.
        Args:
            data (bytes)
            digest_bits (int)
        Returns:
            hex (string)
        """

        if digest_bits == 256:
            khash = sha3.keccak_256()
        else:
            raise NotImplementedError

        khash.update(data)

        return khash.hexdigest()

    @staticmethod
    @deprecated('Unnecessary method')
    def hex2bin(string):
        return bytes.fromhex(string)

    @classmethod
    def hex2bin_recursive(cls, _dict):
        """
        Recursively convert hexdigit dict values to bytes.
        Args:
            _dict (dict)
        Returns:
            dict
        """

        def ctype_xdigit(s):
            """
            Checks if all of the characters in "s" are hexadecimal 'digits'.
            Args:
                s (string): string to check
            """
            return all(c in string.hexdigits for c in s)

        for k, v in _dict.items():
            if type(v) == dict:
                cls.hex2bin_recursive(v)
            elif type(v) == str and ctype_xdigit(v):
                try:
                    _dict[k] = bytes.fromhex(v)
                except ValueError:
                    pass

        return _dict

    @staticmethod
    @deprecated('Unnecessary method')
    def bin2hex(bts):
        return bts.hex()

    @staticmethod
    @deprecated('Unnecessary method')
    def bin2int(number):
        return int.from_bytes(number, 'big')

    @staticmethod
    def get_validator_address(pub_key, upper=True):
        """
        Get validator address from it's pub key (Mp...).
        Validator address is used in signing blocks.
        Args:
            pub_key (string): candidate public key (Mp....)
            upper (bool)
        Returns:
            string, validator address
        """

        pub_key = bytes.fromhex(MinterHelper.prefix_remove(pub_key))
        vaddress = hashlib.sha256(pub_key).hexdigest()[:40]

        return vaddress.upper() if upper else vaddress

    @staticmethod
    def generate_qr(text, fn=None, path='', error='H', version=None, mode=None,
                    output='svg', module_color='black', background='white',
                    quiet_zone=4):
        """
        Generate QR code from text and save to file.
        Detailed documentation for `pyqrcode` package can be found
        here: https://pythonhosted.org/PyQRCode/index.html
        Args:
            text (str): Text, that should be encoded to QR
            fn (str): Filename for generated QR.
                      If not provided random filename is generated.
            path (str): Path to save generate QR
            error (str|int): The error parameter sets the error correction
                             level of the code.
                             Each level has an associated name given by a
                             letter: L, M, Q, or H;
                             each level can correct up to 7, 15, 25, or 30
                             percent of the data respectively.
            version (int): The version parameter specifies the size and data
                           capacity of the code.
                           Versions are any integer between 1 and 40
            mode (str): The mode param sets how the contents will be encoded.
                        Three of the four possible encodings are available.
                        By default, the object uses the most efficient
                        encoding for the contents. You can override this
                        behavior by setting this parameter.
            output (str): Render modes. Available: text|terminal|svg.
                          In `text`|`terminal` modes QR code is printed,
                          `svg` mode saves QR code to file `fn` to path `path`.
            module_color (str): String color of QR code data.
                                Is used only for `terminal` and `svg` modes.
            background (str): String color of QR code background.
                              Is used only for `terminal` and `svg` modes.
            quiet_zone (int): QR code quiet zone.
        Returns:
            fnpath (str): Path to generated QR
        """

        # Generate QR code object
        qrcode = pyqrcode.create(content=text, error=error, version=version,
                                 mode=mode)

        # Render QR code depending on `output` param
        if output == 'text':
            print(qrcode.text(quiet_zone=quiet_zone))
        elif output == 'terminal':
            print(
                qrcode.terminal(
                    module_color=module_color, background=background,
                    quiet_zone=quiet_zone
                )
            )
        elif output == 'svg':
            # Generate filename, if not provided
            if not fn:
                fn = text + str(random.randint(10000, 99999))
                fn = hashlib.sha256(fn.encode()).hexdigest()[:10]
            fnpath = os.path.join(path, fn + '.svg')

            # Save QR code to file
            qrcode.svg(file=fnpath, module_color=module_color,
                       background=background, quiet_zone=quiet_zone)

            return fnpath
        else:
            raise Exception('Wrong QR code render mode')

    @staticmethod
    def bytes_len(value, encoding='utf-8'):
        """
        Count bytes length
        Args:
            value (str|bytes)
            encoding (str)
        """
        if type(value) is str:
            value = bytes(value, encoding=encoding)

        return len(value)

    @staticmethod
    def encode_coin_name(symbol):
        """
        Add nulls to coin name
        Args:
            symbol (string): coin symbol
        Returns:
            string
        """
        return symbol + chr(0) * (10 - len(symbol))

    @staticmethod
    def decode_coin_name(symbol):
        """
        Args:
            symbol (bytes|str)
        Returns:
            string
        """

        if hasattr(symbol, 'decode'):
            symbol = symbol.decode()

        return symbol.replace(chr(0), '')

    @staticmethod
    def to_pip(value):
        """
        Convert BIPs to PIPs.
        Always cast value to str, due to float behaviour:
            Decimal(0.1) = Decimal('0.10000000000004524352345234')
            Decimal('0.1') = Decimal('0.1')
        Args:
            value (str|float|int): value in BIP
        Returns:
            int
        """
        return int(decimal.Decimal(str(value)) * decimal.Decimal(PIP))

    @staticmethod
    def to_bip(value):
        """
        Convert PIPs to BIPs.
        Use dynamic Decimal precision, depending on value length.
        Args:
            value (int|str): value in PIP
        Returns:
            Decimal
        """
        # Check if value is correct PIP value
        if type(value) not in [int, str]:
            raise ValueError(f"{value} should be 'int' or 'str of digits'")
        if type(value) is str and not value.isdigit():
            raise ValueError(f'{value} is not correct PIP value')

        # Get default decimal context
        context = decimal.getcontext()
        # Set temporary decimal context for calculation
        decimal.setcontext(
            decimal.Context(prec=len(str(value)), rounding=decimal.ROUND_DOWN)
        )

        # Convert value
        value = decimal.Decimal(value) / decimal.Decimal(PIP)

        # Reset decimal context to default
        decimal.setcontext(context)

        return value

    @staticmethod
    def prefix_add(value, prefix):
        if prefix not in [PREFIX_ADDR, PREFIX_PUBKEY, PREFIX_CHECK, PREFIX_TX]:
            raise ValueError(f"Unknown prefix '{prefix}'")
        return prefix + value

    @staticmethod
    def prefix_remove(value):
        value = value.replace(PREFIX_ADDR, '')
        value = value.replace(PREFIX_PUBKEY, '')
        value = value.replace(PREFIX_CHECK, '')
        value = value.replace(PREFIX_TX, '')

        return value


@deprecated("Deprecated. Use 'MinterHelper' class instead")
class MinterPrefix:
    """
    Class with minter prefixes and operations with them.
    """

    # Minter wallet address prefix
    ADDRESS = 'Mx'

    # Minter public key prefix
    PUBLIC_KEY = 'Mp'

    # Minter redeem check prefix
    CHECK = 'Mc'

    # Minter transaction prefix
    TRANSACTION = 'Mt'

    @staticmethod
    def remove_prefix(string, prefix):
        return string.replace(prefix, '')



"""
@author: Roman Matusevich
"""
import hashlib

import rlp
from mintersdk import MinterHelper, PREFIX_CHECK, PREFIX_PUBKEY
from mintersdk.sdk import ECDSA
from mintersdk.sdk.wallet import MinterWallet


class MinterCheck(object):
    """
    Minter check class
    Create new check or decode existing
    """

    def __init__(self, nonce, due_block, coin, value, gas_coin,
                 passphrase=None, chain_id=1, **kwargs):
        """
        Args:
            nonce (int)
            due_block (int)
            coin (str)
            value (float)
            gas_coin (str): Gas coin symbol
            passphrase (str)
            chain_id (int)
        """

        self.nonce = nonce
        self.due_block = due_block
        self.coin = coin.upper()
        self.value = value
        self.gas_coin = gas_coin.upper()
        self.passphrase = passphrase
        self.chain_id = chain_id

        for attr, value in kwargs.items():
            setattr(self, attr, value)

    @staticmethod
    def __hash(data):
        """
        Create service hash by RLP encoding and getting keccak hash from
        rlp result
        Args:
            data (list)
        Returns:
            hash (str)
        """
        return MinterHelper.keccak_hash(rlp.encode(data))

    @staticmethod
    def __lockfromsignature(signature):
        """
        Create lock from signature
        Args:
            signature (list): [v, r, s] list
        Returns:
            bytes
        """

        v, r, s = signature
        v = '00' if v == 27 else '01'
        signature = format(r, 'x').zfill(64) + format(s, 'x').zfill(64) + v

        return bytes.fromhex(signature)

    def sign(self, private_key):
        """
        Sign check
        Args:
            private_key (str)
        """

        if not self.passphrase:
            raise ValueError('Passphrase should be not empty string')

        # Prepare structure
        # It contains nonce, chain_id, due_block, coin, value, gas_coin,
        # lock, v, r, s.
        # lock, v, r, s appended later in code
        structure = [
            int(str(self.nonce).encode().hex(), 16),
            self.chain_id,
            self.due_block,
            MinterHelper.encode_coin_name(self.coin),
            MinterHelper.to_pip(self.value),
            MinterHelper.encode_coin_name(self.gas_coin)
        ]

        # Create msg hash
        msg_hash = self.__hash(structure)

        # SHA256 from passphrase
        sha = hashlib.sha256()
        sha.update(self.passphrase.encode())
        passphrase = sha.hexdigest()

        # Create lock from signature
        self.lock = self.__lockfromsignature(
            signature=ECDSA.sign(message=msg_hash, private_key=passphrase)
        )

        # Re-create msg hash with adding lock to structure
        structure.append(self.lock)
        msg_hash = self.__hash(structure)

        # Re-create signature, add it to check attrs and to structure
        signature = ECDSA.sign(message=msg_hash, private_key=private_key)
        self.signature = {
            'v': signature[0],
            'r': format(signature[1], 'x'),
            's': format(signature[2], 'x')
        }
        structure += signature

        # Get RLP, which will be the check
        check = rlp.encode(structure).hex()

        return MinterHelper.prefix_add(check, PREFIX_CHECK)

    @classmethod
    def proof(cls, address, passphrase):
        """
        Create proof
        Args:
            address (str)
            passphrase (str)
        Returns:
            str
        """

        # Get address hash
        address = MinterHelper.prefix_remove(address)
        address = bytes.fromhex(address)
        address_hash = cls.__hash(data=[address])

        # Create SHA256 from passphrase
        sha = hashlib.sha256()
        sha.update(passphrase.encode())
        passphrase = sha.hexdigest()

        # Get signature
        signature = ECDSA.sign(message=address_hash, private_key=passphrase)

        return cls.__lockfromsignature(signature).hex()

    @classmethod
    def from_raw(cls, rawcheck):
        """
        Create check instance from raw check
        Args:
            rawcheck (str)
        Returns:
            MinterCheck
        """

        # Remove check prefix and RLP decode it
        rawcheck = MinterHelper.prefix_remove(rawcheck)
        rawcheck = bytes.fromhex(rawcheck)
        decoded = rlp.decode(rawcheck)

        # Create MinterCheck instance
        kwargs = {
            'nonce': int(decoded[0].decode()),
            'chain_id': int.from_bytes(decoded[1], 'big'),
            'due_block': int.from_bytes(decoded[2], 'big'),
            'coin': MinterHelper.decode_coin_name(decoded[3]),
            'value': MinterHelper.to_bip(int.from_bytes(decoded[4], 'big')),
            'gas_coin': MinterHelper.decode_coin_name(decoded[5]),
            'lock': decoded[6].hex(),
            'signature': {
                'v': int.from_bytes(decoded[7], 'big'),
                'r': decoded[8].hex(),
                's': decoded[9].hex()
            }
        }
        check = MinterCheck(**kwargs)

        # Recover owner address
        msg_hash = cls.__hash(data=[
            int(str(check.nonce).encode().hex(), 16),
            check.chain_id,
            check.due_block,
            MinterHelper.encode_coin_name(check.coin),
            MinterHelper.to_pip(check.value),
            MinterHelper.encode_coin_name(check.gas_coin),
            bytes.fromhex(check.lock)
        ])
        public_key = ECDSA.recover(msg_hash, tuple(check.signature.values()))
        public_key = MinterHelper.prefix_add(public_key, PREFIX_PUBKEY)

        check.owner = MinterWallet.get_address_from_public_key(public_key)

        return check



import unittest
import base64

from mintersdk.sdk.transactions import (
    MinterTx, MinterDelegateTx, MinterSendCoinTx, MinterBuyCoinTx,
    MinterCreateCoinTx, MinterDeclareCandidacyTx, MinterEditCandidateTx,
    MinterRedeemCheckTx, MinterSellAllCoinTx, MinterSellCoinTx,
    MinterSetCandidateOffTx, MinterSetCandidateOnTx, MinterUnbondTx,
    MinterMultiSendCoinTx
)


class TestMinterTx(unittest.TestCase):

    def setUp(self):
        self.SIGNED_TX = 'f8900102018a4d4e540000000000000007b6f5a00eb98ea04ae466d8d38f490db3c99b3996a90e24243952ce9822c6dc1e2c1a438a4d4e5400000000000000888ac7230489e80000808001b845f8431ba01c2c8f702d80cf64da1e9bf1f07a52e2fee8721aebe419aa9f62260a98983f89a07ed297d71d9dc37a57ffe9bb16915dccc703d8c09f30da8aadb9d5dbab8c7da9'
        self.PUBLIC_KEY = 'Mp0eb98ea04ae466d8d38f490db3c99b3996a90e24243952ce9822c6dc1e2c1a43'
        self.TX_FROM = 'Mx9f7fd953c2c69044b901426831ed03ee0bd0597a'
        self.TX_TYPE = 7
        self.TX_STAKE = 10
        self.TX_PAYLOAD = '🔳'  # 4 bytes
        self.tx = MinterTx.from_raw(self.SIGNED_TX)

    def test_instance(self):
        self.assertIsInstance(self.tx, MinterDelegateTx)

    def test_public_key(self):
        self.assertEqual(self.tx.pub_key, self.PUBLIC_KEY)

    def test_from_mx(self):
        self.assertEqual(self.tx.from_mx, self.TX_FROM)

    def test_type(self):
        self.assertEqual(self.tx.type, self.TX_TYPE)

    def test_stake(self):
        self.assertEqual(self.tx.stake, self.TX_STAKE)


class TestMinterBuyCoinTx(unittest.TestCase):

    def setUp(self):
        self.FROM = 'Mx31e61a05adbd13c6b625262704bc305bf7725026'
        self.PRIVATE_KEY = '07bc17abdcee8b971bb8723e36fe9d2523306d5ab2d683631693238e0f9df142'
        self.SIGNED_TX = 'f8830102018a4d4e540000000000000004a9e88a54455354000000000000880de0b6b3a76400008a4d4e5400000000000000880de0b6b3a7640000808001b845f8431ca04ee095a20ca58062a5758e2a6d3941857daa8943b5873c57f111190ca88dbc56a01148bf2fcc721ca353105e4f4a3419bec471d7ae08173f443a28c3ae6d27018a'
        self.TX = MinterBuyCoinTx(**{
            'nonce': 1,
            'chain_id': MinterTx.TESTNET_CHAIN_ID,
            'gas_coin': 'MNT',
            'coin_to_buy': 'TEST',
            'value_to_buy': 1,
            'coin_to_sell': 'MNT',
            'max_value_to_sell': 1
        })

    def test_valid_tx(self):
        """
        Is tx instance of needed TX class.
        """

        self.assertIsInstance(self.TX, MinterBuyCoinTx)

    def test_sign_tx(self):
        """
        Sign transaction and check signed transaction
        """
        self.TX.sign(self.PRIVATE_KEY)

        self.assertEqual(self.TX.signed_tx, self.SIGNED_TX)

    def test_sign_with_signature(self):
        self.TX.signature_type = MinterTx.SIGNATURE_SINGLE_TYPE
        signature = self.TX.generate_signature(self.PRIVATE_KEY)
        self.TX.sign(signature=signature)

        self.assertEqual(self.TX.signed_tx, self.SIGNED_TX)

    def test_from_raw(self):
        tx = MinterTx.from_raw(raw_tx=self.SIGNED_TX)

        self.assertEqual(tx.from_mx, self.FROM)
        self.assertEqual(tx.coin_to_buy, self.TX.coin_to_buy)
        self.assertEqual(tx.value_to_buy, self.TX.value_to_buy)
        self.assertEqual(tx.coin_to_sell, self.TX.coin_to_sell)
        self.assertEqual(tx.max_value_to_sell, self.TX.max_value_to_sell)


class TestMinterCreateCoinTx(unittest.TestCase):

    def setUp(self):
        self.FROM = 'Mx31e61a05adbd13c6b625262704bc305bf7725026'
        self.PRIVATE_KEY = '07bc17abdcee8b971bb8723e36fe9d2523306d5ab2d683631693238e0f9df142'
        self.SIGNED_TX = 'f88f0102018a4d4e540000000000000005b5f48a535550455220544553548a5350525445535400000089056bc75e2d63100000888ac7230489e800000a893635c9adc5dea00000808001b845f8431ca0ccfabd9283d27cf7978bca378e0cc7dc69a39ff3bdc56707fa2d552655f9290da0226057221cbaef35696c9315cd29e783d3c66d842d0a3948a922abb42ca0dabe'
        self.TX = MinterCreateCoinTx(**{
            'nonce': 1,
            'chain_id': MinterTx.TESTNET_CHAIN_ID,
            'gas_coin': 'MNT',
            'name': 'SUPER TEST',
            'symbol': 'SPRTEST',
            'initial_amount': 100,
            'initial_reserve': 10,
            'crr': 10,
            'max_supply': 1000
        })

    def test_valid_tx(self):
        """
        Is tx instance of needed TX class.
        """

        self.assertIsInstance(self.TX, MinterCreateCoinTx)

    def test_sign_tx(self):
        """
        Sign transaction and check signed transaction
        """
        self.TX.sign(self.PRIVATE_KEY)

        self.assertEqual(self.TX.signed_tx, self.SIGNED_TX)

    def test_sign_with_signature(self):
        self.TX.signature_type = MinterTx.SIGNATURE_SINGLE_TYPE
        signature = self.TX.generate_signature(self.PRIVATE_KEY)
        self.TX.sign(signature=signature)

        self.assertEqual(self.TX.signed_tx, self.SIGNED_TX)

    def test_from_raw(self):
        tx = MinterTx.from_raw(raw_tx=self.SIGNED_TX)

        self.assertEqual(tx.from_mx, self.FROM)
        self.assertEqual(tx.name, self.TX.name)
        self.assertEqual(tx.symbol, self.TX.symbol)
        self.assertEqual(tx.initial_amount, self.TX.initial_amount)
        self.assertEqual(tx.initial_reserve, self.TX.initial_reserve)
        self.assertEqual(tx.crr, self.TX.crr)
        self.assertEqual(tx.max_supply, self.TX.max_supply)


class TestMinterDeclareCandidacyTx(unittest.TestCase):

    def setUp(self):
        self.PRIVATE_KEY = '6e1df6ec69638d152f563c5eca6c13cdb5db4055861efc11ec1cdd578afd96bf'
        self.SIGNED_TX = 'f8a80102018a4d4e540000000000000006b84df84b949f7fd953c2c69044b901426831ed03ee0bd0597aa00eb98ea04ae466d8d38f490db3c99b3996a90e24243952ce9822c6dc1e2c1a430a8a4d4e5400000000000000884563918244f40000808001b845f8431ca0c379230cbe09103b31983402c9138ad29d839bcecee70e11ac9bf5cfe70850d9a06c92bfb9a627bfaefc3ad46fc60ff1fdc42efe0e8805d57f20795a403c91e8bd'
        self.TX = MinterDeclareCandidacyTx(**{
            'nonce': 1,
            'chain_id': MinterTx.TESTNET_CHAIN_ID,
            'gas_coin': 'MNT',
            'address': 'Mx9f7fd953c2c69044b901426831ed03ee0bd0597a',
            'pub_key': 'Mp0eb98ea04ae466d8d38f490db3c99b3996a90e24243952ce9822c6dc1e2c1a43',
            'commission': 10,
            'coin': 'MNT',
            'stake': 5
        })

    def test_valid_tx(self):
        """
        Is tx instance of needed TX class.
        """

        self.assertIsInstance(self.TX, MinterDeclareCandidacyTx)

    def test_sign_tx(self):
        """
        Sign transaction and check signed transaction
        """
        self.TX.sign(self.PRIVATE_KEY)

        self.assertEqual(self.TX.signed_tx, self.SIGNED_TX)

    def test_sign_with_signature(self):
        self.TX.signature_type = MinterTx.SIGNATURE_SINGLE_TYPE
        signature = self.TX.generate_signature(self.PRIVATE_KEY)
        self.TX.sign(signature=signature)

        self.assertEqual(self.TX.signed_tx, self.SIGNED_TX)

    def test_from_raw(self):
        tx = MinterTx.from_raw(raw_tx=self.SIGNED_TX)

        self.assertEqual(tx.from_mx, self.TX.address)
        self.assertEqual(tx.address, self.TX.address)
        self.assertEqual(tx.pub_key, self.TX.pub_key)
        self.assertEqual(tx.commission, self.TX.commission)
        self.assertEqual(tx.coin, self.TX.coin)
        self.assertEqual(tx.stake, self.TX.stake)


class TestMinterDelegateTx(unittest.TestCase):

    def setUp(self):
        self.PRIVATE_KEY = '6e1df6ec69638d152f563c5eca6c13cdb5db4055861efc11ec1cdd578afd96bf'
        self.PUBLIC_KEY = 'Mp0eb98ea04ae466d8d38f490db3c99b3996a90e24243952ce9822c6dc1e2c1a43'
        self.SIGNED_TX = 'f8900102018a4d4e540000000000000007b6f5a00eb98ea04ae466d8d38f490db3c99b3996a90e24243952ce9822c6dc1e2c1a438a4d4e5400000000000000888ac7230489e80000808001b845f8431ba01c2c8f702d80cf64da1e9bf1f07a52e2fee8721aebe419aa9f62260a98983f89a07ed297d71d9dc37a57ffe9bb16915dccc703d8c09f30da8aadb9d5dbab8c7da9'

        self.TX = MinterDelegateTx(**{
            'nonce': 1,
            'chain_id': MinterTx.TESTNET_CHAIN_ID,
            'gas_coin': 'MNT',
            'pub_key': self.PUBLIC_KEY,
            'coin': 'MNT',
            'stake': 10
        })

    def test_valid_tx(self):
        """
        Is tx instance of needed TX class.
        """

        self.assertIsInstance(self.TX, MinterDelegateTx)

    def test_sign_tx(self):
        """
        Sign transaction and check signed transaction
        """
        self.TX.sign(self.PRIVATE_KEY)

        self.assertEqual(self.TX.signed_tx, self.SIGNED_TX)

    def test_sign_with_signature(self):
        self.TX.signature_type = MinterTx.SIGNATURE_SINGLE_TYPE
        signature = self.TX.generate_signature(self.PRIVATE_KEY)
        self.TX.sign(signature=signature)

        self.assertEqual(self.TX.signed_tx, self.SIGNED_TX)

    def test_from_raw(self):
        tx = MinterTx.from_raw(raw_tx=self.SIGNED_TX)

        self.assertEqual(tx.pub_key, self.PUBLIC_KEY)
        self.assertEqual(tx.coin, self.TX.coin)
        self.assertEqual(tx.stake, self.TX.stake)


class TestMinterRedeemCheckTx(unittest.TestCase):

    def setUp(self):
        self.PRIVATE_KEY = '05ddcd4e6f7d248ed1388f0091fe345bf9bf4fc2390384e26005e7675c98b3c1'
        self.SIGNED_TX = 'f9013f0102018a4d4e540000000000000009b8e4f8e2b89df89b01830f423f8a4d4e5400000000000000843b9aca00b8419b3beac2c6ad88a8bd54d24912754bb820e58345731cb1b9bc0885ee74f9e50a58a80aa990a29c98b05541b266af99d3825bb1e5ed4e540c6e2f7c9b40af9ecc011ca00f7ba6d0aa47d74274b960fba02be03158d0374b978dcaa5f56fc7cf1754f821a019a829a3b7bba2fc290f5c96e469851a3876376d6a6a4df937327b3a5e9e8297b841da021d4f84728e0d3d312a18ec84c21768e0caa12a53cb0a1452771f72b0d1a91770ae139fd6c23bcf8cec50f5f2e733eabb8482cf29ee540e56c6639aac469600808001b845f8431ba009493b3296a085a27f2bc015ad5c1cc644ba21bdce1b78a49e987227f24a87a3a07187da48b6ea528d372ed33923f5d74011f56cc2db3cab2cf5b4bbab97990373'
        self.TX = MinterRedeemCheckTx(**{
            'nonce': 1,
            'chain_id': MinterTx.TESTNET_CHAIN_ID,
            'gas_coin': 'MNT',
            'check': 'Mcf89b01830f423f8a4d4e5400000000000000843b9aca00b8419b3beac2c6ad88a8bd54d24912754bb820e58345731cb1b9bc0885ee74f9e50a58a80aa990a29c98b05541b266af99d3825bb1e5ed4e540c6e2f7c9b40af9ecc011ca00f7ba6d0aa47d74274b960fba02be03158d0374b978dcaa5f56fc7cf1754f821a019a829a3b7bba2fc290f5c96e469851a3876376d6a6a4df937327b3a5e9e8297',
            'proof': 'da021d4f84728e0d3d312a18ec84c21768e0caa12a53cb0a1452771f72b0d1a91770ae139fd6c23bcf8cec50f5f2e733eabb8482cf29ee540e56c6639aac469600'
        })

    def test_valid_tx(self):
        """
        Is tx instance of needed TX class.
        """

        self.assertIsInstance(self.TX, MinterRedeemCheckTx)

    def test_sign_tx(self):
        """
        Sign transaction and check signed transaction
        """
        self.TX.sign(self.PRIVATE_KEY)

        self.assertEqual(self.TX.signed_tx, self.SIGNED_TX)

    def test_sign_with_signature(self):
        self.TX.signature_type = MinterTx.SIGNATURE_SINGLE_TYPE
        signature = self.TX.generate_signature(self.PRIVATE_KEY)
        self.TX.sign(signature=signature)

        self.assertEqual(self.TX.signed_tx, self.SIGNED_TX)

    def test_from_raw(self):
        tx = MinterTx.from_raw(raw_tx=self.SIGNED_TX)

        self.assertEqual(tx.check, self.TX.check)
        self.assertEqual(tx.proof, self.TX.proof)


class TestMinterSellAllCoinTx(unittest.TestCase):

    def setUp(self):
        self.FROM = 'Mx31e61a05adbd13c6b625262704bc305bf7725026'
        self.PRIVATE_KEY = '07bc17abdcee8b971bb8723e36fe9d2523306d5ab2d683631693238e0f9df142'
        self.SIGNED_TX = 'f87a0102018a4d4e540000000000000003a0df8a4d4e54000000000000008a54455354000000000000880de0b6b3a7640000808001b845f8431ca0b10794a196b6ad2f94e6162613ca9538429dd49ca493594ba9d99f80d2499765a03c1d78e9e04f57336691e8812a16faccb00bf92ac817ab61cd9bf001e9380d47'
        self.TX = MinterSellAllCoinTx(**{
            'nonce': 1,
            'chain_id': MinterTx.TESTNET_CHAIN_ID,
            'gas_coin': 'MNT',
            'coin_to_sell': 'MNT',
            'coin_to_buy': 'TEST',
            'min_value_to_buy': 1
        })

    def test_valid_tx(self):
        """
        Is tx instance of needed TX class.
        """

        self.assertIsInstance(self.TX, MinterSellAllCoinTx)

    def test_sign_tx(self):
        """
        Sign transaction and check signed transaction
        """
        self.TX.sign(self.PRIVATE_KEY)

        self.assertEqual(self.TX.signed_tx, self.SIGNED_TX)

    def test_sign_with_signature(self):
        self.TX.signature_type = MinterTx.SIGNATURE_SINGLE_TYPE
        signature = self.TX.generate_signature(self.PRIVATE_KEY)
        self.TX.sign(signature=signature)

        self.assertEqual(self.TX.signed_tx, self.SIGNED_TX)

    def test_from_raw(self):
        tx = MinterTx.from_raw(raw_tx=self.SIGNED_TX)

        self.assertEqual(tx.from_mx, self.FROM)
        self.assertEqual(tx.coin_to_sell, self.TX.coin_to_sell)
        self.assertEqual(tx.coin_to_buy, self.TX.coin_to_buy)
        self.assertEqual(tx.min_value_to_buy, self.TX.min_value_to_buy)


class TestMinterSellCoinTx(unittest.TestCase):

    def setUp(self):
        self.FROM = 'Mx31e61a05adbd13c6b625262704bc305bf7725026'
        self.PRIVATE_KEY = '07bc17abdcee8b971bb8723e36fe9d2523306d5ab2d683631693238e0f9df142'
        self.SIGNED_TX = 'f8830102018a4d4e540000000000000002a9e88a4d4e5400000000000000880de0b6b3a76400008a54455354000000000000880de0b6b3a7640000808001b845f8431ba0e34be907a18acb5a1aed263ef419f32f5adc6e772b92f949906b497bba557df3a0291d7704980994f7a6f5950ca84720746b5928f21c3cfc5a5fbca2a9f4d35db0'
        self.TX = MinterSellCoinTx(**{
            'nonce': 1,
            'chain_id': MinterTx.TESTNET_CHAIN_ID,
            'gas_coin': 'MNT',
            'coin_to_sell': 'MNT',
            'value_to_sell': 1,
            'coin_to_buy': 'TEST',
            'min_value_to_buy': 1
        })

    def test_valid_tx(self):
        """
        Is tx instance of needed TX class.
        """

        self.assertIsInstance(self.TX, MinterSellCoinTx)

    def test_sign_tx(self):
        """
        Sign transaction and check signed transaction
        """
        self.TX.sign(self.PRIVATE_KEY)

        self.assertEqual(self.TX.signed_tx, self.SIGNED_TX)

    def test_sign_with_signature(self):
        self.TX.signature_type = MinterTx.SIGNATURE_SINGLE_TYPE
        signature = self.TX.generate_signature(self.PRIVATE_KEY)
        self.TX.sign(signature=signature)

        self.assertEqual(self.TX.signed_tx, self.SIGNED_TX)

    def test_from_raw(self):
        tx = MinterTx.from_raw(raw_tx=self.SIGNED_TX)

        self.assertEqual(tx.from_mx, self.FROM)
        self.assertEqual(tx.coin_to_sell, self.TX.coin_to_sell)
        self.assertEqual(tx.value_to_sell, self.TX.value_to_sell)
        self.assertEqual(tx.coin_to_buy, self.TX.coin_to_buy)
        self.assertEqual(tx.min_value_to_buy, self.TX.min_value_to_buy)


class TestMinterSendTx(unittest.TestCase):

    def setUp(self):
        self.PRIVATE_KEY = '07bc17abdcee8b971bb8723e36fe9d2523306d5ab2d683631693238e0f9df142'
        self.TO = 'Mx1b685a7c1e78726c48f619c497a07ed75fe00483'
        self.FROM = 'Mx31e61a05adbd13c6b625262704bc305bf7725026'
        self.SIGNED_TX = 'f8840102018a4d4e540000000000000001aae98a4d4e5400000000000000941b685a7c1e78726c48f619c497a07ed75fe00483880de0b6b3a7640000808001b845f8431ca01f36e51600baa1d89d2bee64def9ac5d88c518cdefe45e3de66a3cf9fe410de4a01bc2228dc419a97ded0efe6848de906fbe6c659092167ef0e7dcb8d15024123a'
        self.TX = MinterSendCoinTx(**{
            'nonce': 1,
            'chain_id': MinterTx.TESTNET_CHAIN_ID,
            'gas_coin': 'MNT',
            'to': self.TO,
            'coin': 'MNT',
            'value': 1
        })

    def test_valid_tx(self):
        """
        Is tx instance of needed TX class.
        """

        self.assertIsInstance(self.TX, MinterSendCoinTx)

    def test_sign_tx(self):
        """
        Sign transaction and check signed transaction
        """
        self.TX.sign(self.PRIVATE_KEY)

        self.assertEqual(self.TX.signed_tx, self.SIGNED_TX)

    def test_sign_with_signature(self):
        self.TX.signature_type = MinterTx.SIGNATURE_SINGLE_TYPE
        signature = self.TX.generate_signature(self.PRIVATE_KEY)
        self.TX.sign(signature=signature)

        self.assertEqual(self.TX.signed_tx, self.SIGNED_TX)

    def test_from_raw(self):
        tx = MinterTx.from_raw(self.SIGNED_TX)

        self.assertEqual(tx.from_mx, self.FROM)
        self.assertEqual(tx.to, self.TX.to)
        self.assertEqual(tx.coin, self.TX.coin)
        self.assertEqual(tx.value, self.TX.value)


class TestMinterSetCandidateOffTx(unittest.TestCase):

    def setUp(self):
        self.PRIVATE_KEY = '05ddcd4e6f7d248ed1388f0091fe345bf9bf4fc2390384e26005e7675c98b3c1'
        self.SIGNED_TX = 'f87c0102018a4d4e54000000000000000ba2e1a00eb98ea04ae466d8d38f490db3c99b3996a90e24243952ce9822c6dc1e2c1a43808001b845f8431ca02ac45817f167c34b55b8afa0b6d9692be28e2aa41dd28a134663d1f5bebb5ad8a06d5f161a625701d506db20c497d24e9939c2e342a6ff7d724cb1962267bd4ba5'
        self.TX = MinterSetCandidateOffTx(**{
            'nonce': 1,
            'chain_id': MinterTx.TESTNET_CHAIN_ID,
            'gas_coin': 'MNT',
            'pub_key': 'Mp0eb98ea04ae466d8d38f490db3c99b3996a90e24243952ce9822c6dc1e2c1a43'
        })

    def test_valid_tx(self):
        """
        Is tx instance of needed TX class.
        """

        self.assertIsInstance(self.TX, MinterSetCandidateOffTx)

    def test_sign_tx(self):
        """
        Sign transaction and check signed transaction
        """
        self.TX.sign(self.PRIVATE_KEY)

        self.assertEqual(self.TX.signed_tx, self.SIGNED_TX)

    def test_sign_with_signature(self):
        self.TX.signature_type = MinterTx.SIGNATURE_SINGLE_TYPE
        signature = self.TX.generate_signature(self.PRIVATE_KEY)
        self.TX.sign(signature=signature)

        self.assertEqual(self.TX.signed_tx, self.SIGNED_TX)

    def test_from_raw(self):
        tx = MinterTx.from_raw(raw_tx=self.SIGNED_TX)

        self.assertEqual(tx.pub_key, self.TX.pub_key)


class TestMinterSetCandidateOnTx(unittest.TestCase):

    def setUp(self):
        self.PRIVATE_KEY = '05ddcd4e6f7d248ed1388f0091fe345bf9bf4fc2390384e26005e7675c98b3c1'
        self.SIGNED_TX = 'f87c0102018a4d4e54000000000000000aa2e1a00eb98ea04ae466d8d38f490db3c99b3996a90e24243952ce9822c6dc1e2c1a43808001b845f8431ba0095aed433171fe5ac385ccd299507bdcad3dd2269794fd0d14d4f58327ddc87ea046ec7e4f8f9b477a1255485f36e0567e62283723ecc5a0bd1e5d201e53e85245'
        self.TX = MinterSetCandidateOnTx(**{
            'nonce': 1,
            'chain_id': MinterTx.TESTNET_CHAIN_ID,
            'gas_coin': 'MNT',
            'pub_key': 'Mp0eb98ea04ae466d8d38f490db3c99b3996a90e24243952ce9822c6dc1e2c1a43'
        })

    def test_valid_tx(self):
        """
        Is tx instance of needed TX class.
        """

        self.assertIsInstance(self.TX, MinterSetCandidateOnTx)

    def test_sign_tx(self):
        """
        Sign transaction and check signed transaction
        """
        self.TX.sign(self.PRIVATE_KEY)

        self.assertEqual(self.TX.signed_tx, self.SIGNED_TX)

    def test_sign_with_signature(self):
        self.TX.signature_type = MinterTx.SIGNATURE_SINGLE_TYPE
        signature = self.TX.generate_signature(self.PRIVATE_KEY)
        self.TX.sign(signature=signature)

        self.assertEqual(self.TX.signed_tx, self.SIGNED_TX)

    def test_from_raw(self):
        tx = MinterTx.from_raw(raw_tx=self.SIGNED_TX)

        self.assertEqual(tx.pub_key, self.TX.pub_key)


class TestMinterUnbondTx(unittest.TestCase):

    def setUp(self):
        self.PRIVATE_KEY = '6e1df6ec69638d152f563c5eca6c13cdb5db4055861efc11ec1cdd578afd96bf'
        self.SIGNED_TX = 'f88f0102018a4d4e540000000000000008b6f5a00eb98ea04ae466d8d38f490db3c99b3996a90e24243952ce9822c6dc1e2c1a438a4d4e5400000000000000888ac7230489e80000808001b844f8421ca0ff5766c85847b37a276f3f9d027fb7c99745920fa395c7bd399cedd8265c5e1d9f791bcdfe4d1bc1e73ada7bf833103c828f22d83189dad2b22ad28a54aacf2a'
        self.TX = MinterUnbondTx(**{
            'nonce': 1,
            'chain_id': MinterTx.TESTNET_CHAIN_ID,
            'gas_coin': 'MNT',
            'pub_key': 'Mp0eb98ea04ae466d8d38f490db3c99b3996a90e24243952ce9822c6dc1e2c1a43',
            'coin': 'MNT',
            'value': 10
        })

    def test_valid_tx(self):
        """
        Is tx instance of needed TX class.
        """

        self.assertIsInstance(self.TX, MinterUnbondTx)

    def test_sign_tx(self):
        """
        Sign transaction and check signed transaction
        """
        self.TX.sign(self.PRIVATE_KEY)

        self.assertEqual(self.TX.signed_tx, self.SIGNED_TX)

    def test_sign_with_signature(self):
        self.TX.signature_type = MinterTx.SIGNATURE_SINGLE_TYPE
        signature = self.TX.generate_signature(self.PRIVATE_KEY)
        self.TX.sign(signature=signature)

        self.assertEqual(self.TX.signed_tx, self.SIGNED_TX)

    def test_from_raw(self):
        tx = MinterTx.from_raw(raw_tx=self.SIGNED_TX)

        self.assertEqual(tx.pub_key, self.TX.pub_key)
        self.assertEqual(tx.coin, self.TX.coin)
        self.assertEqual(tx.value, self.TX.value)


class TestMinterEditCandidateTx(unittest.TestCase):

    def setUp(self):
        self.FROM = 'Mxa879439b0a29ecc7c5a0afe54b9eb3c22dbde8d9'
        self.PRIVATE_KEY = 'a3fb55450f53dbbf4f2494280188f7f0cd51a7b51ec27ed49ed364d920e326ba'
        self.SIGNED_TX = 'f8a80102018a4d4e54000000000000000eb84df84ba04ae1ee73e6136c85b0ca933a9a1347758a334885f10b3238398a67ac2eb153b89489e5dc185e6bab772ac8e00cf3fb3f4cb0931c4794e731fcddd37bb6e72286597d22516c8ba3ddffa0808001b845f8431ca0421470f27f78231b669c1bf1fcc56168954d64fbb7dc3ff021bab01311fab6eaa075e86365d98c87e806fcbc5c542792f569e19d8ae7af671d9ba4679acc86d35e'
        self.TX = MinterEditCandidateTx(**{
            'nonce': 1,
            'chain_id': MinterTx.TESTNET_CHAIN_ID,
            'gas_coin': 'MNT',
            'pub_key': 'Mp4ae1ee73e6136c85b0ca933a9a1347758a334885f10b3238398a67ac2eb153b8',
            'reward_address': 'Mx89e5dc185e6bab772ac8e00cf3fb3f4cb0931c47',
            'owner_address': 'Mxe731fcddd37bb6e72286597d22516c8ba3ddffa0'
        })

    def test_valid_tx(self):
        """
        Is tx instance of needed TX class.
        """

        self.assertIsInstance(self.TX, MinterEditCandidateTx)

    def test_sign_tx(self):
        """
        Sign transaction and check signed transaction
        """
        self.TX.sign(self.PRIVATE_KEY)

        self.assertEqual(self.TX.signed_tx, self.SIGNED_TX)

    def test_sign_with_signature(self):
        self.TX.signature_type = MinterTx.SIGNATURE_SINGLE_TYPE
        signature = self.TX.generate_signature(self.PRIVATE_KEY)
        self.TX.sign(signature=signature)

        self.assertEqual(self.TX.signed_tx, self.SIGNED_TX)

    def test_from_raw(self):
        tx = MinterTx.from_raw(raw_tx=self.SIGNED_TX)

        self.assertEqual(tx.from_mx, self.FROM)
        self.assertEqual(tx.pub_key, self.TX.pub_key)
        self.assertEqual(tx.reward_address, self.TX.reward_address)
        self.assertEqual(tx.owner_address, self.TX.owner_address)


class TestMinterMultiSendCoinTx(unittest.TestCase):

    def setUp(self):
        self.FROM = 'Mx31e61a05adbd13c6b625262704bc305bf7725026'
        self.PRIVATE_KEY = '07bc17abdcee8b971bb8723e36fe9d2523306d5ab2d683631693238e0f9df142'
        self.SIGNED_TX = 'f8b30102018a4d4e54000000000000000db858f856f854e98a4d4e540000000000000094fe60014a6e9ac91618f5d1cab3fd58cded61ee9988016345785d8a0000e98a4d4e540000000000000094ddab6281766ad86497741ff91b6b48fe85012e3c8802c68af0bb140000808001b845f8431ca0b15dcf2e013df1a2aea02e36a17af266d8ee129cdcb3e881d15b70c9457e7571a0226af7bdaca9d42d6774c100b22e0c7ba4ec8dd664d17986318e905613013283'
        self.TX = MinterMultiSendCoinTx(**{
            'nonce': 1,
            'chain_id': MinterTx.TESTNET_CHAIN_ID,
            'gas_coin': 'MNT',
            'txs': [
                {
                    'coin': 'MNT',
                    'to': 'Mxfe60014a6e9ac91618f5d1cab3fd58cded61ee99',
                    'value': 0.1
                },
                {
                    'coin': 'MNT',
                    'to': 'Mxddab6281766ad86497741ff91b6b48fe85012e3c',
                    'value': 0.2
                }
            ]
        })

    def test_valid_tx(self):
        """
        Is tx instance of needed TX class.
        """

        self.assertIsInstance(self.TX, MinterMultiSendCoinTx)

    def test_sign_tx(self):
        """
        Sign transaction and check signed transaction
        """
        self.TX.sign(self.PRIVATE_KEY)

        self.assertEqual(self.TX.signed_tx, self.SIGNED_TX)

    def test_from_raw(self):
        tx = MinterTx.from_raw(raw_tx=self.SIGNED_TX)

        self.assertEqual(tx.from_mx, self.FROM)

    def test_sign_with_signature(self):
        self.TX.signature_type = MinterTx.SIGNATURE_SINGLE_TYPE
        signature = self.TX.generate_signature(self.PRIVATE_KEY)
        self.TX.sign(signature=signature)

        self.assertEqual(self.TX.signed_tx, self.SIGNED_TX)


class TestTxFees(unittest.TestCase):

    def setUp(self):
        self.TO = 'Mx31e61a05adbd13c6b625262704bc305bf7725026'
        self.PRIVATE_KEY = '07bc17abdcee8b971bb8723e36fe9d2523306d5ab2d683631693238e0f9df142'
        self.TX_PAYLOAD_UTF = '🔳'  # 4 bytes
        self.EXPECTED_SEND_COIN_FEE = 18000000000000000

        self.MULTISEND_RECIPIENTS = [
            {
                'coin': 'MNT',
                'to': 'Mxfe60014a6e9ac91618f5d1cab3fd58cded61ee99',
                'value': 0.1
            },
            {
                'coin': 'MNT',
                'to': 'Mxddab6281766ad86497741ff91b6b48fe85012e3c',
                'value': 0.2
            }
        ]
        self.EXPECTED_MULTISEND_FEE = 15000000000000000

    def test_payload_fee(self):
        tx = MinterSendCoinTx(**{
            'nonce': 1,
            'chain_id': MinterTx.TESTNET_CHAIN_ID,
            'gas_coin': 'MNT',
            'to': self.TO,
            'coin': 'MNT',
            'value': 1,
            'payload': self.TX_PAYLOAD_UTF
        })
        tx.sign(self.PRIVATE_KEY)
        actual_fee = tx.get_fee()
        self.assertEqual(self.EXPECTED_SEND_COIN_FEE, actual_fee)

    def test_multisend_fee(self):
        tx = MinterMultiSendCoinTx(**{
            'nonce': 1,
            'chain_id': MinterTx.TESTNET_CHAIN_ID,
            'gas_coin': 'MNT',
            'txs': self.MULTISEND_RECIPIENTS
        })
        tx.sign(self.PRIVATE_KEY)
        actual_fee = tx.get_fee()
        self.assertEqual(self.EXPECTED_MULTISEND_FEE, actual_fee)


class TestMinterSendMultisigTx(unittest.TestCase):

    def setUp(self):
        self.PRIVATE_KEYS = [
            'b354c3d1d456d5a1ddd65ca05fd710117701ec69d82dac1858986049a0385af9',
            '38b7dfb77426247aed6081f769ed8f62aaec2ee2b38336110ac4f7484478dccb',
            '94c0915734f92dd66acfdc48f82b1d0b208efd544fe763386160ec30c968b4af'
        ]
        self.TO = 'Mxd82558ea00eb81d35f2654953598f5d51737d31d'
        self.FROM = 'Mxdb4f4b6942cb927e8d7e3a1f602d0f1fb43b5bd2'
        self.SIGNED_TX = 'f901270102018a4d4e540000000000000001aae98a4d4e540000000000000094d82558ea00eb81d35f2654953598f5d51737d31d880de0b6b3a7640000808002b8e8f8e694db4f4b6942cb927e8d7e3a1f602d0f1fb43b5bd2f8cff8431ca0a116e33d2fea86a213577fc9dae16a7e4cadb375499f378b33cddd1d4113b6c1a021ee1e9eb61bbd24233a0967e1c745ab23001cf8816bb217d01ed4595c6cb2cdf8431ca0f7f9c7a6734ab2db210356161f2d012aa9936ee506d88d8d0cba15ad6c84f8a7a04b71b87cbbe7905942de839211daa984325a15bdeca6eea75e5d0f28f9aaeef8f8431ba0d8c640d7605034eefc8870a6a3d1c22e2f589a9319288342632b1c4e6ce35128a055fe3f93f31044033fe7b07963d547ac50bccaac38a057ce61665374c72fb454'
        self.TX = MinterSendCoinTx(**{
            'nonce': 1,
            'chain_id': MinterTx.TESTNET_CHAIN_ID,
            'gas_coin': 'MNT',
            'to': self.TO,
            'coin': 'MNT',
            'value': 1
        })

    def test_valid_tx(self):
        """
        Is tx instance of needed TX class.
        """

        self.assertIsInstance(self.TX, MinterSendCoinTx)

    def test_sign_tx(self):
        """
        Sign transaction and check signed transaction
        """
        self.TX.sign(private_key=self.PRIVATE_KEYS, ms_address=self.FROM)

        self.assertEqual(self.TX.signed_tx, self.SIGNED_TX)

    def test_from_raw(self):
        tx = MinterTx.from_raw(self.SIGNED_TX)

        self.assertEqual(tx.from_mx, self.FROM)
        self.assertEqual(tx.to, self.TX.to)
        self.assertEqual(tx.coin, self.TX.coin)
        self.assertEqual(tx.value, self.TX.value)

    def test_add_signature(self):
        # Sign tx with 2 of 3 private keys
        self.TX.sign(private_key=self.PRIVATE_KEYS[:2], ms_address=self.FROM)
        # Add signature by 3rd private key
        self.TX = MinterTx.add_signature(self.TX.signed_tx, self.PRIVATE_KEYS[2])

        self.assertEqual(self.TX.signed_tx, self.SIGNED_TX)

    def test_sign_with_signature(self):
        # Set signature type for transaction
        self.TX.signature_type = MinterTx.SIGNATURE_MULTI_TYPE

        # Generate signatures
        signatures = []
        for pk in self.PRIVATE_KEYS:
            signature = self.TX.generate_signature(private_key=pk)
            signatures.append(signature)

        # Sign transaction with signatures
        self.TX.sign(signature=signatures, ms_address=self.FROM)

        self.assertEqual(self.TX.signed_tx, self.SIGNED_TX)

    def test_sign_with_pk_and_signature(self):
        # Set signature type for transaction
        self.TX.signature_type = MinterTx.SIGNATURE_MULTI_TYPE

        # Generate 1 signature
        signatures = []
        for pk in self.PRIVATE_KEYS:
            signatures.append(self.TX.generate_signature(private_key=pk))

        # Sign transaction with pks and signature
        self.TX.sign(private_key=self.PRIVATE_KEYS[:2], signature=signatures[2], ms_address=self.FROM)
        self.assertEqual(self.TX.signed_tx, self.SIGNED_TX)

        self.TX.sign(private_key=self.PRIVATE_KEYS[0], signature=signatures[1:], ms_address=self.FROM)
        self.assertEqual(self.TX.signed_tx, self.SIGNED_TX)


class TestPayloadsFromRaw(unittest.TestCase):
    def setUp(self):
        self.TO = 'Mxd82558ea00eb81d35f2654953598f5d51737d31d'
        self.FROM = 'Mx31e61a05adbd13c6b625262704bc305bf7725026'
        self.PK = '07bc17abdcee8b971bb8723e36fe9d2523306d5ab2d683631693238e0f9df142'
        self.TX = MinterSendCoinTx(
            nonce=1, gas_coin='mnt', to=self.TO, coin='mnt', value=1
        )
        self.TX_DECODED = None

    def sign_and_decode(self, payload):
        self.TX.payload = payload
        self.TX.sign(private_key=self.PK)
        self.TX_DECODED = MinterTx.from_raw(self.TX.signed_tx)

    def test_hex_like(self):
        payload = 'fff'
        self.sign_and_decode(payload)

        self.assertEqual(payload, self.TX_DECODED.payload)
        self.assertEqual(self.FROM, self.TX_DECODED.from_mx)

    def test_str_bytes(self):
        payload = '🔳'
        self.sign_and_decode(payload)

        self.assertEqual(payload, self.TX_DECODED.payload)
        self.assertEqual(self.FROM, self.TX_DECODED.from_mx)

    def test_raw_bytes(self):
        payload = b'\xff\xff\xff'
        self.sign_and_decode(payload)

        self.assertEqual(payload, self.TX_DECODED.payload)
        self.assertEqual(self.FROM, self.TX_DECODED.from_mx)


class TestFromBase64(unittest.TestCase):
    def setUp(self):
        self.B64_TXS = [
            '+Hw8AQGKQklQAAAAAAAAAAKi4YpPTEJJVFgAAAAAiQHluPqP4qwAAIpMRU1PTgAAAAAAgICAAbhF+EMboPtT5w1Brh5BO66qs75e1sOj9Ka4KfxGQDOFsQssgNn/oAeDnTkaMj685tdWvWa6rUmViaCB+KerPBDHUE7O731j',
            '+H+DBJaHAQGKQklQAAAAAAAAAAGi4YpCSVAAAAAAAAAAlJXK8ve/qfgqz5QEbX52KCQKOQ/bgICAAbhF+EMcoFVh1TOlkwUHmSpEvOQprRnZgnpSeIFxXn15fApbl28qoFR5GhbH34BiCmpwqY+Qs4xI5DjfE9PNfVYEZ7axgyvc',
            '+IeCGxkBAYpCSVAAAAAAAAAAAavqikNPTlNVTEdBTUWUm7rikGga0Jtdjuse5KRJp2SOn8WJAaBVaQ2duAAAgIABuEX4QxygBQd9EAxqcsKigdqsvCEVA5GapLPxdlbZ/DYkC3RXpYOgJ2QuKIW1U/yrTVec56v06V42VwaO2VRqGvuJVLkojvU=',
            '+HqCMGYBAYpCSVAAAAAAAAAAA57dikJJUAAAAAAAAACKQ0VOVEFVUlVTAIZa8xB6QACAgAG4RfhDG6A+PinAE4fNMpPwC8U8/DbHNSIERcWDE9rridQ1DyECAaBiGkXNmAgUnYK2VjJjMLJGWLN4T3jz/2KylTFZqXNM5Q==',
            '+H6CBYcBAYpCSVAAAAAAAAAAAqLhikJJUAAAAAAAAACJARWORgkT0AAAilRBUFRBUAAAAACAgIABuEX4Qxug8tXe0wJ61FNJq+p/KEVsTE044Cq5mqtnD55xyUM2niigCPjkKq+K5Rabeghgyf7+zxjOhykruz0dQOoZR5GzbZ8=',
            '+JKB2AEBikJJUAAAAAAAAAAHt/agd/cYNBCOm15lI3o5JjYxtPmanVhDehOFyTDBPuHU4qaKT05MWTEAAAAAAIm+S9/Lh9P4ZoqAgAG4RfhDG6AVkWCf6F/LQlIUjBFQFCcfAgWxhRBTs7ZdLDLOaUQntaB/C/kHw/fRLDH/ZMM5hdZpx87pZZfTRTuQ6Peda7ParA==',
            '+IRVAQGKQklQAAAAAAAAAAGq6YpCSVAAAAAAAAAAlG3tXZ5JAX3uqMMYO7aKsxx9JGTxiIrHIwSJ6AAAgIABuEX4QxugLtA0N8YgapJtAF4/oq4mgnDeTAFy64tAfhlY2POh4fmgTB1pdMhcBzpSiPqUTVPTCn+xwAxpTD8eQsV5+ZWEM40=',
            '+JKCTDIBAYpCSVAAAAAAAAAAB7b1oGKbVSjwnRx0qD0YQU8uQmPhSFDEej+sP4VfIAERERERikJJUAAAAAAAAACIRWORgkT0AACAgAG4RfhDG6BbXtr6FKf4Ifdlji3nCSqUQ3OEfReYBUfa71+zNb0wh6AEZSWMY7CGXvh3Rqjjiam9IT3NL7gvI/wrqxSrFkxj/g==',
            '+JOCB6IBAYpCSVAAAAAAAAAAB7f2oGKbVSjwnRx0qD0YQU8uQmPhSFDEej+sP4VfIAERERERikJJUAAAAAAAAACJFa8deLWMQAAAgIABuEX4Qxyg3MYUWNZZPy1k4ikM97HJ7RJuLCHyteqUymPVQ+r+qJygcN2yKmUK9adEaqjQRNEHesgRE+dAlamHu2Bd7OE4zwI=',
            '+JOCAZIBAYpCSVAAAAAAAAAAB7f2oGKbVSjwnRx0qD0YQU8uQmPhSFDEej+sP4VfIAERERERikJJUAAAAAAAAACJBbEq76+oBAAAgIABuEX4QxygLl0jQ98RtdXCkIKDYce9OD11tpA4kmAtPLXIGkeycAKgGRsYDVVx+SD98N4zKByD2H+w0yToFuZk8gN2PtWVNpQ=',
            '+IQJAQGKQklQAAAAAAAAAAGq6YpCSVAAAAAAAAAAlJ9/UFrwbYh8dflHt3kRDT99fUG/iJNP9bP1XNAAgIABuEX4QxugOu3JcCce9OO5rpWDVsQpGk+gwTLbt3p8qu2YM7Tg1UGgF8Ez00im24HKFlJM2DWoUSJ3G1BjzREg1AGG6dP1s1s=',
            '+H6CB1UBAYpCSVAAAAAAAAAAAqLhikJJUAAAAAAAAACJEENWGogpMAAAikNFTlRBVVJVUwCAgIABuEX4QxugfLaJkvWttDTqA9EelJ+9RQ8anWwngbIkeOuQydh0RI2gB0JC+oxPrSmB8dZmazeh3ot2Ff6bE2czGfwgPu7ZpCg=',
            '+PaCAw8BAYpCSVAAAAAAAAAAAavqikJJUAAAAAAAAACUvW+bnucw5qCqmK4SHaLeJ1gp4YSJYadP9azWWcAAuG5CaXBleC5uZXQgLSBTdWNjZXNzIEJpcCB3aXRoZHJhd2FsICMxMzczOC4gT3VyIGN1cnJlbnQgZXhjaGFuZ2UgcmF0ZXMgZm9yIDEgQklQOiAwLjAxNDckIC8gMC4wMTUzJC4gRm9sbG93IHVzLoABuEX4Qxug7zYujmw36ehtj+scZxn+Zlpj3NrtUp924DwBS9+hkJagQtz8zLrB4jgzLnFtDv1F0t1gKQnpU6PyTYGuicMMRe0=',
            '+H+DBJaLAQGKQklQAAAAAAAAAAGi4YpCSVAAAAAAAAAAlJXK8ve/qfgqz5QEbX52KCQKOQ/bgICAAbhF+EMboKGqlP2cC2tqtsnntOBRAg9XNsVBBb5wLBjCPlp5RuBaoDLYDXbDwKQz/ONyYoDV9AUI0C7wCME+T/gGeR1BEdCu',
            '+JOCARABAYpCSVAAAAAAAAAAB7f2oEiBrRZ8pftYhjIoQfmS1ortiU/8tYq8CA6K07FW8QRbikJJUAAAAAAAAACJA9NIdC2o2QDmgIABuEX4QxugGdqhQ+dkzUft+Pc5/A0TYUXcZLGDLA8C/5Q2GzNhRoygaAxGtwGHh9VCrOUVsUCvxOyMSZkMqIbC2i7LVKNY6qU=',
            '+IQJAQGKQklQAAAAAAAAAAGq6YpCSVAAAAAAAAAAlJ9/UFrwbYh8dflHt3kRDT99fUG/iJb2P7XLwAAAgIABuEX4QxygHiZ64dA9EZbsmjLMtDWPvs/Sn4vLjTeCjAizxgEyFhegG9HDwfYDX249KO/kIVXnOsUUBTA1u6KvsImtRTYwyA0=',
            '+H6CB1YBAYpCSVAAAAAAAAAAAqLhikJJUAAAAAAAAACJEENWGogpMAAAik9ORUJJUAAAAACAgIABuEX4Qxug81K8K87BiQpmawv8wBT/8wJA7xexSiStKlH+ijGE+/agCOLB7aqTVi62WHw2BgRBIlYMICc3LOSqsWE6lFku3Ew=',
            '+IaCAwQBAYpCSVAAAAAAAAAAAarpikZBTENPTgAAAACUhfwWxlBErQ8ZAvfDq2FnB0UsxeaIDeC2s6dkAACAgAG4RfhDG6C1m77cTSA1qllKv8FOW3/BaIO3lVQtrVclDUd5ay3NpaBPHg/nkVgAfN63rRKBYiNZBb+4sa4tVjSLlL2v6yVB5w==',
            '+IMrAQGKQklQAAAAAAAAAAKp6IpCSVAAAAAAAAAAiAr2pNB8jwAAikVDT05BAAAAAACITCl3L4h8OpGAgAG4RfhDG6DX7yY6y9xkV640fWT9EMcIH/WDrPu3Y4hX9zpESaijPKBVP7Z56G2J1urB3QZq9cl7AaYbE7EZeAMWsCsidQGXvg==',
            '+IUBAQGKQklQAAAAAAAAAAGr6opCSVAAAAAAAAAAlK/SZxtm+xqSGATRe5jj+eCa+bhmiXPriL12u+CcAICAAbhF+EMboJioZmk/kgs9vAcJcvM8VwmoRuws0s02eMT9yMqNr5s7oBk5m7YSzSIKDjRQYRjyaktm+zMQI4JluYnQVSOpbVkt'
        ]

    @staticmethod
    def base64ToHex(b64str):
        b64_bytes = b64str.encode()
        tx_bytes = base64.b64decode(b64_bytes)

        return tx_bytes.hex()

    def test_txs(self):
        for index, b64_tx in enumerate(self.B64_TXS):
            try:
                raw_tx = self.base64ToHex(b64_tx)
                MinterTx.from_raw(raw_tx)
            except Exception as e:
                self.fail(f'Tx #{index} from base64 failed: {e.__str__()}')


if __name__ == '__main__':
    unittest.main()



"""
@author: Roman Matusevich
"""
import sslcrypto


class ECDSA:
    """
    ECDSA class
    """

    # Curve data
    curve = sslcrypto.ecc.get_curve('secp256k1')
    pub_key_len = curve._backend.public_key_length

    @classmethod
    def sign(cls, message, private_key):
        """
        Args:
            message (string): message to sign
            private_key (string): private_key
        Returns:
            list(int)
        """

        # Create signature
        signature = cls.curve.sign(
            data=bytes.fromhex(message), private_key=bytes.fromhex(private_key),
            recoverable=True, hash=None
        )
        v = signature[0]
        r = int.from_bytes(signature[1:cls.pub_key_len+1], byteorder='big')
        s = int.from_bytes(signature[cls.pub_key_len+1:], byteorder='big')

        return [v, r, s]

    @classmethod
    def recover(cls, message, vrs):
        """
        Args:
            message (string): message
            vrs (tuple): tuple of v, r, s (r, s in hex)
        Returns:
            str
        """

        # Create signature
        signature = (
            vrs[0].to_bytes(length=1, byteorder='big') +
            int(vrs[1], 16).to_bytes(
                length=cls.pub_key_len, byteorder='big'
            ) +
            int(vrs[2], 16).to_bytes(
                length=cls.pub_key_len, byteorder='big'
            )
        )

        # Get raw recover of public key.
        pub_key_raw = cls.curve.recover(
            signature=signature, data=bytes.fromhex(message), hash=None
        )

        # Convert public key to hex electrum format
        x, y = cls.curve.decode_public_key(pub_key_raw)
        pub_key = x.hex() + y.hex()

        return pub_key



import unittest

from mintersdk.sdk.wallet import MinterWallet


class TestMinterWallet(unittest.TestCase):

    def setUp(self):
        self.mnemonic = 'slice better asset talent state citizen dry maze base agent source reveal'
        self.private_key = '7ffc6bc08f2d8a0ead1d3f64e6a9862b7695dafceca24f25978341447594aa07'
        self.address = 'Mx5a4c6c7fbd05ff8e5b09818db5ad229852784e01'

    def test_private_key(self):
        wallet = MinterWallet.create(mnemonic=self.mnemonic)
        self.assertEqual(wallet['private_key'], self.private_key)

    def test_address(self):
        wallet = MinterWallet.create(mnemonic=self.mnemonic)
        self.assertEqual(wallet['address'], self.address)

    def test_creation(self):
        for _ in range(250):
            MinterWallet.create()



import unittest

from mintersdk.sdk.transactions import MinterSendCoinTx
from mintersdk.sdk.deeplink import MinterDeeplink


class TestDeeplink(unittest.TestCase):
    def test_fulltx(self):
        tx = MinterSendCoinTx(
            coin='BIP', to='Mx18467bbb64a8edf890201d526c35957d82be3d95',
            value=1.23456789, nonce=1, gas_coin='MNT', gas_price=1,
            payload='Check payload'
        )
        deeplink = MinterDeeplink(tx=tx)

        self.assertEqual(
            MinterDeeplink.BASE_URL + '/-EcBqumKQklQAAAAAAAAAJQYRnu7ZKjt-JAgHVJsNZV9gr49lYgRIhD0do20AI1DaGVjayBwYXlsb2FkAQGKTU5UAAAAAAAAAA',
            deeplink.generate()
        )

    def test_data_part_with_payload(self):
        tx = MinterSendCoinTx(
            coin='MNT', to='Mx7633980c000139dd3bd24a3f54e06474fa941e16',
            value=10, nonce=1, gas_coin='ASD', gas_price=1,
            payload='custom message'
        )
        deeplink = MinterDeeplink(tx=tx)
        deeplink.nonce = ''
        deeplink.gas_price = ''

        self.assertEqual(
            MinterDeeplink.BASE_URL + '/-EgBqumKTU5UAAAAAAAAAJR2M5gMAAE53TvSSj9U4GR0-pQeFoiKxyMEiegAAI5jdXN0b20gbWVzc2FnZYCAikFTRAAAAAAAAAA',
            deeplink.generate()
        )

    def test_data_only(self):
        tx = MinterSendCoinTx(
            coin='BIP', to='Mx18467bbb64a8edf890201d526c35957d82be3d95',
            value=1.23456789, nonce=1, gas_coin='MNT', gas_price=1,
            payload='Hello World'
        )
        deeplink = MinterDeeplink(tx=tx, data_only=True)

        self.assertEqual(
            MinterDeeplink.BASE_URL + '/8AGq6YpCSVAAAAAAAAAAlBhGe7tkqO34kCAdUmw1lX2Cvj2ViBEiEPR2jbQAgICAgA',
            deeplink.generate()
        )



"""
@author: Roman Matusevich
"""
import base64

import rlp
from mintersdk import MinterHelper


class MinterDeeplink(object):
    """ Create deeplink for Minter transaction """

    BASE_URL = 'https://bip.to/tx'

    def __init__(self, tx, data_only=False, base_url=BASE_URL):
        """
        Create deeplink object
        Args:
            tx (MinterTx): MinterTx object
            data_only (bool): Generate deeplink only with tx data
            base_url (str): Base URL for generated deeplink
        """
        super(MinterDeeplink, self).__init__()

        self.base_url = base_url

        # Generate deeplink structure attributes
        self.__type = tx.TYPE
        self.__data = self.__get_tx_data(tx=tx)
        self.nonce = tx.nonce if not data_only else ''
        self.gas_price = tx.gas_price if not data_only else ''
        self.gas_coin = tx.gas_coin if not data_only else ''
        self.payload = tx.payload if not data_only else ''

    @staticmethod
    def __get_tx_data(tx):
        """ Get data from transaction """
        structure = tx._structure_from_instance()
        return rlp.encode(list(structure['data'].values()))

    def generate(self, password=None):
        """
        Generate deeplink
        Args:
            password (str): Check password
        Returns:
            deeplink (str)
        """

        # Create deeplink structure
        gas_coin = MinterHelper.encode_coin_name(
            self.gas_coin
        ) if self.gas_coin else ''
        deep_structure = [self.__type, self.__data, self.payload, self.nonce,
                          self.gas_price, gas_coin]

        # Create deephash base64 urlsafe
        deephash = rlp.encode(deep_structure)
        deephash = base64.urlsafe_b64encode(deephash)
        deephash = deephash.decode().rstrip('=')

        # Create deeplink URL
        deeplink = self.base_url + '/' + deephash

        # If password check needed, add (`p` URL param)
        if password:
            password = password.encode().hex()
            password = base64.urlsafe_b64encode(password.encode())
            password = password.decode().rstrip('=')

            deeplink += '?p=' + password

        return deeplink



"""
@author: Roman
"""
import requests
import json

from deprecated import deprecated
from mintersdk import MinterHelper


class MinterAPI(object):
    """
    Base MinterAPI class
    """
    # API host
    api_url = ''

    # Timeout connecting to host
    connect_timeout = 1

    # Timeout reading from host
    read_timeout = 3

    # Default request headers
    headers = {
      'Content-Type': 'application/json'
    }

    def __init__(self, api_url, **kwargs):
        """
        Args:
            api_url (str): API host, e.g. http://localhost/api/
            kwargs: Any other attributes you need
                    Predefined kwargs:
                        - connect_timeout (float|int)
                        - read_timeout (float|int)
                        - headers (dict)

        """
        self.api_url = api_url
        if self.api_url[-1] != '/':
            self.api_url += '/'

        for name, value in kwargs.items():
            setattr(self, name, value)

    def get_status(self):
        """ Get node status """
        return self._request(command='status')

    def get_candidate(self, public_key, height=None, pip2bip=False):
        """
        Get candidate
        Args:
            public_key (string): candidate public key
            height (int): block height,
            pip2bip (bool): Convert coin amounts to BIP (default is in PIP)
        """
        response = self._request(
            command='candidate',
            params={'pub_key': public_key, 'height': height}
        )

        if pip2bip:
            return self.__response_processor(
                data=response,
                funcs=[(self.__pip_to_bip, {'exclude': ['commission']})],
            )

        return response

    def get_validators(self, height=None, page=None, limit=None):
        """
        Get validators list
        Args:
            height (int): get validators on specified block height
            page (int|None): page number
            limit (int|None): items per page
        """
        return self._request(
            command='validators',
            params={'height': height, 'page': page, 'perPage': limit}
        )

    def get_addresses(self, addresses, height=None, pip2bip=False):
        """
        Returns addresses balances
        Args:
            addresses (list[str]): Addresses list
            height (int|None): Block height
            pip2bip (bool): Convert coin amounts to BIP (default is in PIP)
        """
        response = self._request(
            command='addresses',
            params={'addresses': json.dumps(addresses), 'height': height}
        )

        if pip2bip:
            return self.__response_processor(
                data=response, funcs=[self.__pip_to_bip]
            )

        return response

    def get_balance(self, address, height=None, pip2bip=False):
        """
        Get balance by address
        Args:
            address (string): wallet address
            height (int): block height
            pip2bip (bool): Convert coin amounts to BIP (default is in PIP)
        """
        response = self._request(
            command='address', params={'address': address, 'height': height}
        )

        if pip2bip:
            return self.__response_processor(
                data=response, funcs=[self.__pip_to_bip]
            )

        return response

    def get_nonce(self, address):
        """
        Nonce - int, used for prevent transaction reply
        Args:
            address (string): wallet address
        """

        balance = self.get_balance(address)
        nonce = balance['result']['transaction_count'] + 1

        return nonce

    def send_transaction(self, tx):
        """
        Send transaction
        Args:
            tx (string): signed transaction to send
        """
        return self._request(
            command='send_transaction', params={'tx': '0x' + tx}
        )

    def get_transaction(self, tx_hash, pip2bip=False):
        """
        Get transaction info
        Args:
            tx_hash (string): transaction hash
            pip2bip (bool): Convert coin amounts to BIP (default is in PIP)
        """
        response = self._request(
            command='transaction', params={'hash': '0x' + tx_hash}
        )

        if pip2bip:
            return self.__response_processor(
                data=response,
                funcs=[(self.__pip_to_bip, {'exclude': ['commission']})]
            )

        return response

    def get_block(self, height, pip2bip=False):
        """
        Get block data at given height
        Args:
            height (int): block height
            pip2bip (bool): Convert coin amounts to BIP (default is in PIP)
        """
        response = self._request(command='block', params={'height': height})

        if pip2bip:
            return self.__response_processor(
                data=response, funcs=[self.__pip_to_bip]
            )

        return response

    def get_latest_block_height(self):
        """
        Get latest block height
        """
        return self.get_status()['result']['latest_block_height']

    def get_events(self, height, pip2bip=False):
        """
        Get events at given height
        Args:
            height (int): block height
            pip2bip (bool): Convert coin amounts to BIP (default is in PIP)
        """
        response = self._request(command='events', params={'height': height})

        if pip2bip:
            return self.__response_processor(
                data=response, funcs=[self.__pip_to_bip]
            )

        return response

    def get_candidates(self, height=None, include_stakes=False, pip2bip=False):
        """
        Get candidates
        Args:
            height (int): block height
            include_stakes (bool)
            pip2bip (bool): Convert coin amounts to BIP (default is in PIP)
        """
        response = self._request(
            command='candidates',
            params={
                'height': height,
                'include_stakes': str(include_stakes).lower()
            }
        )

        if pip2bip:
            return self.__response_processor(
                data=response,
                funcs=[(self.__pip_to_bip, {'exclude': ['commission']})]
            )

        return response

    def get_coin_info(self, symbol, height=None, pip2bip=False):
        """
        Get information about coin
        Args:
            symbol (string): coin name
            height (int): block height
            pip2bip (bool): Convert coin amounts to BIP (default is in PIP)
        """
        response = self._request(
            command='coin_info',
            params={'symbol': symbol.upper(), 'height': height}
        )

        if pip2bip:
            return self.__response_processor(
                data=response, funcs=[self.__pip_to_bip]
            )

        return response

    def estimate_coin_sell(self, coin_to_sell, value_to_sell, coin_to_buy,
                           height=None, pip2bip=False):
        """
        Return estimate of sell coin transaction
        Args:
            coin_to_sell (string): coin name to sell
            value_to_sell (string|int): Amount of coins to sell in PIP.
                    Provide `value_to_sell` in PIP, if `pip2bip` False.
                    Provide `value_to_sell` in BIP, if `pip2bip` True.
            coin_to_buy (string): coin name to buy
            height (int): block height
            pip2bip (bool): Convert coin amounts to BIP (default is in PIP)
        """
        # Convert `value_to_sell` to PIP, if needed
        if pip2bip:
            value_to_sell = MinterHelper.to_pip(value_to_sell)

        # Get default response
        response = self._request(
            command='estimate_coin_sell',
            params={
                'coin_to_sell': coin_to_sell.upper(),
                'value_to_sell': value_to_sell,
                'coin_to_buy': coin_to_buy.upper(),
                'height': height
            }
        )

        # Convert response values from PIP to BIP, if needed
        if pip2bip:
            return self.__response_processor(
                data=response, funcs=[self.__pip_to_bip]
            )

        return response

    def estimate_coin_sell_all(self, coin_to_sell, value_to_sell, coin_to_buy,
                               height=None, pip2bip=False):
        """
        Return estimate of sell all coin transaction.
        Args:
            coin_to_sell (string): coin name to sell
            value_to_sell (string|int): Amount of coins to sell in PIP.
                    Provide `value_to_sell` in PIP, if `pip2bip` False.
                    Provide `value_to_sell` in BIP, if `pip2bip` True.
            coin_to_buy (string): coin name to buy
            height (int): block height
            pip2bip (bool): Convert coin amounts to BIP (default is in PIP)
        """
        # Convert `value_to_sell` to PIP, if needed
        if pip2bip:
            value_to_sell = MinterHelper.to_pip(value_to_sell)

        # Get default response
        response = self._request(
            command='estimate_coin_sell_all',
            params={
                'coin_to_sell': coin_to_sell.upper(),
                'value_to_sell': value_to_sell,
                'coin_to_buy': coin_to_buy.upper(),
                'height': height
            }
        )

        # Convert response values from PIP to BIP, if needed
        if pip2bip:
            return self.__response_processor(
                data=response, funcs=[self.__pip_to_bip]
            )

        return response

    def estimate_coin_buy(self, coin_to_sell, value_to_buy, coin_to_buy,
                          height=None, pip2bip=False):
        """
        Return estimate of buy coin transaction
        Args:
            coin_to_sell (string): coin name to sell
            value_to_buy (string): Amount of coins to buy in PIP.
                    Provide `value_to_buy` in PIP, if `pip2bip` False.
                    Provide `value_to_buy` in BIP, if `pip2bip` True.
            coin_to_buy (string): coin name to buy
            height (int): block height
            pip2bip (bool): Convert coin amounts to BIP (default is in PIP)
        """
        # Convert `value_to_buy` to PIP, if needed
        if pip2bip:
            value_to_buy = MinterHelper.to_pip(value_to_buy)

        # Get default response
        response = self._request(
            command='estimate_coin_buy',
            params={
                'coin_to_sell': coin_to_sell,
                'value_to_buy': value_to_buy,
                'coin_to_buy': coin_to_buy,
                'height': height
            }
        )

        # Convert response values from PIP to BIP, if needed
        if pip2bip:
            return self.__response_processor(
                data=response, funcs=[self.__pip_to_bip]
            )

        return response

    @deprecated("Please, use 'estimate_tx_commission' instead")
    def estimate_tx_comission(self, tx, height=None):
        """
        Estimate current tx gas.
        Args:
            tx (string): signed transaction
            height (int|None): block height
        """
        if tx[:2] != '0x':
            tx = '0x' + tx

        return self._request(
            command='estimate_tx_commission',
            params={'tx': tx, 'height': height}
        )

    def estimate_tx_commission(self, tx, height=None, pip2bip=False):
        """
        Estimate current tx gas.
        Args:
            tx (string): signed transaction
            height (int|None): block height
            pip2bip (bool): Convert coin amounts to BIP (default is in PIP)
        """
        if tx[:2] != '0x':
            tx = '0x' + tx

        response = self._request(
            command='estimate_tx_commission',
            params={'tx': tx, 'height': height}
        )

        if pip2bip:
            return self.__response_processor(
                data=response, funcs=[self.__pip_to_bip]
            )

        return response

    def get_transactions(self, query, page=None, limit=None, pip2bip=False):
        """
        Get transactions by query.
        Args:
            query (string)
            page (int)
            limit (int)
            pip2bip (bool): Convert coin amounts to BIP (default is in PIP)
        """
        response = self._request(
            command='transactions',
            params={'query': query, 'page': page, 'perPage': limit}
        )

        if pip2bip:
            return self.__response_processor(
                data=response, funcs=[self.__pip_to_bip]
            )

        return response

    def get_unconfirmed_transactions(self, limit=None):
        """
        Get unconfirmed transactions.
        Args:
            limit (int)
        """
        return self._request(
            command='unconfirmed_txs', params={'limit': limit}
        )

    def get_max_gas_price(self, height=None):
        """
        Returns current max gas price.
        Args:
            height (int)
        """
        return self._request(command='max_gas', params={'height': height})

    def get_min_gas_price(self):
        """
        Returns min gas price.
        """
        return self._request(command='min_gas_price')

    def get_missed_blocks(self, public_key, height=None):
        """
        Returns missed blocks by validator public key.
        Args:
            public_key (str): candidate public key
            height (int): block chain height
        """
        return self._request(
            command='missed_blocks',
            params={'pub_key': public_key, 'height': height}
        )

    def get_genesis(self, pip2bip=False):
        """
        Return network genesis.
        Args:
            pip2bip (bool): Convert coin amounts to BIP (default is in PIP)
        """
        response = self._request(command='genesis')
        if pip2bip:
            return self.__response_processor(
                data=response, funcs=[self.__pip_to_bip]
            )
        return response

    def get_network_info(self):
        """ Return node network information. """
        return self._request(command='net_info')

    def _request(self, command, request_type='get', **kwargs):
        """
        Send all requests to API
        Args:
            command (str): API command
            request_type (str): Request type (GET|POST)
            kwargs: requests package arguments
        """
        # Add timeouts if were not set
        if not kwargs.get('timeout', None):
            kwargs['timeout'] = (self.connect_timeout, self.read_timeout)

        # Add headers
        if not kwargs.get('headers', None):
            kwargs['headers'] = self.headers

        # Trying make request
        try:
            url = self.api_url + command

            if request_type == 'get':
                response = requests.get(url, **kwargs)
            elif request_type == 'post':
                response = requests.post(url, **kwargs)
            else:
                response = None

            # Try to get json response and prepare result
            try:
                return self.__response_processor(
                    data=response.json(), funcs=[self.__digits_to_int]
                )
            except Exception as e:
                msg = 'Response parse JSON error: {}; Response is: {}'
                raise Exception(msg.format(e.__str__(), response.text))
        except requests.exceptions.ReadTimeout:
            raise
        except requests.exceptions.ConnectTimeout:
            raise
        except requests.exceptions.ConnectionError:
            raise
        except requests.exceptions.HTTPError:
            raise
        except ValueError:
            raise

    @staticmethod
    def __digits_to_int(value, key, exclude=None):
        """
        Numeric strings to integers converter.
        Used as processor function for '__response_processor()'
        Args:
            exclude (list|None): keys to be excluded
        Returns:
            int|any
        """
        # Combine default exclude list with argument
        exclude_keys = ['tx.type']
        if type(exclude) is list:
            exclude_keys += exclude

        # Convert value
        if type(value) is str and value.isdigit() and \
                (key not in exclude_keys or key is None):
            return int(value)

        return value

    @staticmethod
    def __pip_to_bip(value, key, exclude=None):
        """
        Convert coin amounts integers from PIP to BIP.
        Used as processor function for '__response_processor()'
        Args:
            exclude ([str]): Keys excluded from conversion
        Returns:
            Decimal|int
        """
        # Keys with coin values. Keys' values should be converted
        # from PIP to BIP
        include_keys = [
            'total_stake', 'value', 'bip_value', 'value_to_buy', 'volume',
            'maximum_value_to_sell', 'initial_amount', 'initial_reserve',
            'max_supply', 'stake', 'value_to_sell', 'tx.sell_amount',
            'minimum_value_to_buy', 'tx.return', 'block_reward', 'amount',
            'reserve_balance', 'will_get', 'commission', 'total_bip_stake',
            'will_pay', 'accum_reward', 'total_slashed'
        ]

        # Combine default exclude list with argument
        exclude_keys = []
        if type(exclude) is list:
            exclude_keys += exclude

        # Convert value.
        # Key should be present in 'include_keys' and should be missing
        # in 'exclude_keys' or key is coin symbol.
        # DANGEROUS! We determine if key is coin symbol by checking
        # if key is uppercase or not (should think more about it).
        if type(value) is int and \
                (
                    (key in include_keys and key not in exclude_keys) or
                    key.isupper()
                ):
            return MinterHelper.to_bip(value)

        return value

    @staticmethod
    def __response_processor(data, funcs):
        """
        Works only with successful response: response status is 200
        and contains 'result' key.
        Args:
            data (dict): Response text dict (json parsed)
            funcs (list): Functions to be applied to dict values
        Returns:
            dict
        """

        def data_recursive(result, fn):
            """
            Method is used for looping response 'result' until plain
            values are got.
            Then this values are processed by 'fn'.
            Args:
                result (any): 'result' key value from response text
                fn (function|[function, kwargs]): function to process value
            Returns:
                result (any)
            """
            # Unpack fn if list or tuple
            processor = fn
            kwargs = {}
            if type(fn) in [list, tuple]:
                processor, kwargs = fn

            # Go recursion
            if type(result) is list:
                for item in result:
                    data_recursive(item, fn)
            elif type(result) is dict:
                for key, value in result.items():
                    if type(value) in [list, dict]:
                        data_recursive(value, fn)
                    else:
                        result[key] = processor(value, key, **kwargs)
            else:
                return processor(value=result, key=None, **kwargs)

            return result

        # If there is no 'result' key in data, return original data
        if not data.get('result'):
            return data

        # Otherwise apply value processor functions to 'result' key values
        for func in funcs:
            data['result'] = data_recursive(result=data['result'], fn=func)

        return data



"""
@author: Roman Matusevich
"""
import rlp
import hashlib
import copy
from mintersdk import (
    MinterHelper, PREFIX_ADDR, PREFIX_TX, PREFIX_PUBKEY, PREFIX_CHECK
)
from mintersdk.sdk import ECDSA
from mintersdk.sdk.wallet import MinterWallet


class MinterTx(object):
    """
    Base transaction class.
    Used only for inheritance by real transaction classes.
    """

    # Fee in PIP
    PAYLOAD_COMMISSION = 2

    # All gas price multiplied by FEE DEFAULT (PIP)
    FEE_DEFAULT_MULTIPLIER = 1000000000000000

    # Type of single signature for the transaction
    SIGNATURE_SINGLE_TYPE = 1

    # Type of multi signature for the transaction
    SIGNATURE_MULTI_TYPE = 2

    # Main net chain id
    MAINNET_CHAIN_ID = 1

    # Test net chain id
    TESTNET_CHAIN_ID = 2

    # Each minter transaction has:
    # Nonce - int, used for prevent transaction reply.
    # Gas Price - big int, used for managing transaction fees.
    # Gas Coin - 10 bytes, symbol of a coin to pay fee
    # Type - type of transaction (is not needed for base tx class)
    # Data - data of transaction (depends on transaction type).
    # Payload (arbitrary bytes) - arbitrary user-defined bytes.
    # Service Data - reserved field.
    # Signature Type - single or multisig transaction.
    # Signature Data - digital signature of transaction.
    # Don't change this dict directly. You need to copy this dict
    # and make needed changes.
    _STRUCTURE_DICT = {
        'nonce': None,
        'chain_id': None,
        'gas_price': None,
        'gas_coin': None,
        'type': None,
        'data': None,
        'payload': '',
        'service_data': '',
        'signature_type': None,
        'signature_data': ''
    }

    def __init__(self, nonce, gas_coin, payload='', service_data='',
                 chain_id=1, gas_price=1, **kwargs):
        if self.__class__ is MinterTx:
            exc_msg = """You can not directly create instance of MinterTx.
            Please use one of subclasses ({}) to create needed transaction."""
            raise Exception(exc_msg.format(self.__class__.__subclasses__()))

        # Set every tx attributes
        self.nonce = nonce
        self.chain_id = chain_id
        self.gas_coin = gas_coin.upper()
        self.gas_price = gas_price
        self.payload = payload
        self.service_data = service_data
        self.signature_type = None
        self.signed_tx = None

        for name, value in kwargs.items():
            setattr(self, name, value)

    def generate_tx_rlp(self):
        """
        Create structure from instance and prepare rlp structure
        Returns:
            tx_struct (dict)
        """
        # Get structure populated with instance data
        tx_struct = self._structure_from_instance()
        # Remove signature data, it's not needed before getting Keccak
        tx_struct.pop('signature_data')

        # Encode tx data to RLP
        tx_struct['data'] = rlp.encode(list(tx_struct['data'].values()))

        return tx_struct

    @staticmethod
    def generate_signed_tx(tx_struct):
        """
        Generate signed tx hash from it's structure
        Args:
            tx_struct (dict): populated tx structure dict
        Returns:
            signed_tx hash (str)
        """
        tx_rlp = rlp.encode(list(tx_struct.values()))
        return tx_rlp.hex()

    def generate_signature(self, private_key):
        """
        Create signature for transaction
        Args:
            private_key (str): private key to sign with
        Returns:
            hex_signature (str)
        """
        # Get structure populated with instance data and rlp encoded
        tx_struct = self.generate_tx_rlp()

        # Create keccak hash
        tx_rlp = rlp.encode(list(tx_struct.values()))
        keccak = MinterHelper.keccak_hash(tx_rlp)

        # Create signature
        signature = ECDSA.sign(keccak, private_key)
        signature = rlp.encode(signature).hex()

        return signature

    def sign(self, private_key=None, signature=None, ms_address=None):
        """
        Sign transaction.
        This method can be called only from instances of inherited
        classes.
        Args:
            private_key (string|list[string]): private key to sign with
            signature (string|list[string]): signature to sign with
            ms_address (string): Multi signature address to sign tx by
        """
        # Check arguments validity
        if not private_key and not signature:
            raise Exception(
                'Please, provide either `private_key(s)` or `signature(s)`'
            )
        if not ms_address and private_key and type(private_key) is not str:
            raise Exception(
                '''
                Please, provide a single `private_key` or set `ms_address` 
                argument for multisig tx
                '''
            )
        if not ms_address and signature and type(signature) is not str:
            raise Exception(
                '''
                Please, provide a single `signature` or set `ms_address` 
                argument for multisig tx
                '''
            )

        # Set tx signature type
        self.signature_type = self.SIGNATURE_SINGLE_TYPE
        if ms_address:
            if type(private_key) is str:
                private_key = [private_key]
            if type(signature) is str:
                signature = [signature]
            self.signature_type = self.SIGNATURE_MULTI_TYPE

        # Get populated and rlp encoded tx structure
        tx_struct = self.generate_tx_rlp()

        # Signature data
        if self.signature_type == self.SIGNATURE_SINGLE_TYPE:
            # Only one of private_key or signature can be provided for
            # single signature type tx
            if private_key and signature:
                raise Exception(
                    '''
                    Please, provide one of `private_key` or `signature` for 
                    single signature type tx
                    '''
                )

            # Set signature_data
            if private_key:
                signature_data = self.generate_signature(private_key)
            else:
                signature_data = signature
            signature_data = self.decode_signature(signature_data)
        else:
            # Add multisig address to signature
            signature_data = [
                bytes.fromhex(MinterHelper.prefix_remove(ms_address)),
                []
            ]

            # Sign by each private key and add to total signature data
            if private_key:
                for pk in private_key:
                    _signature = self.generate_signature(pk)
                    signature_data[1].append(self.decode_signature(_signature))

            # Sign by each signature and add to total signature data
            if signature:
                for _signature in signature:
                    signature_data[1].append(self.decode_signature(_signature))
        tx_struct['signature_data'] = rlp.encode(signature_data)

        self.signed_tx = self.generate_signed_tx(tx_struct)

    def get_hash(self):
        """
        Generate tx hash with prefix
        Returns:
            string
        """
        if not hasattr(self, 'signed_tx') or not self.signed_tx:
            raise AttributeError('You need to sign transaction before')

        # Create SHA256
        sha = hashlib.sha256()
        sha.update(bytes.fromhex(self.signed_tx))

        # Return first 40 symbols with prefix
        return MinterHelper.prefix_add(sha.hexdigest()[:40], PREFIX_TX)

    def _structure_from_instance(self):
        """
        Populating structure dict by instance values.
        Prepare values for signing process.
        """

        struct = copy.copy(self._STRUCTURE_DICT)

        struct.update({
            'nonce': self.nonce,
            'chain_id': self.chain_id,
            'gas_price': self.gas_price,
            'gas_coin': MinterHelper.encode_coin_name(self.gas_coin),
            'payload': self.payload,
            'service_data': self.service_data,
            'signature_type': self.signature_type
        })

        return struct

    @classmethod
    def _structure_to_kwargs(cls, structure):
        """
        Works with already populated structure and prepare **kwargs for
        creating new instance of tx.
        """

        structure.update({
            'gas_coin': MinterHelper.decode_coin_name(structure['gas_coin'])
        })

        return structure

    def get_fee(self):
        """
        Get fee of transaction in PIP.
        Returns:
            int
        """
        # Commission for payload and service_data bytes
        payload_gas = (
            MinterHelper.bytes_len(self.payload) * self.PAYLOAD_COMMISSION
        )
        service_data_gas = (
            MinterHelper.bytes_len(self.service_data) * self.PAYLOAD_COMMISSION
        )

        # Total commission
        commission = self.COMMISSION + payload_gas + service_data_gas
        commission *= self.FEE_DEFAULT_MULTIPLIER

        return commission

    @classmethod
    def from_raw(cls, raw_tx):
        """
        Generate tx object from raw tx
        Args:
            raw_tx (string)
        Returns:
            MinterTx child instance
        """

        tx = rlp.decode(bytes.fromhex(raw_tx))

        # Try to decode payload
        try:
            payload = tx[6].decode()
        except UnicodeDecodeError:
            payload = tx[6]

        # Try to decode service data
        try:
            service_data = tx[7].decode()
        except UnicodeDecodeError:
            service_data = tx[7]

        # Populate structure dict with decoded tx data
        struct = copy.copy(cls._STRUCTURE_DICT)
        struct.update({
            'nonce': int.from_bytes(tx[0], 'big'),
            'chain_id': int.from_bytes(tx[1], 'big'),
            'gas_price': int.from_bytes(tx[2], 'big'),
            'gas_coin': tx[3].decode(),
            'type': int.from_bytes(tx[4], 'big'),
            'payload': payload,
            'service_data': service_data,
            'signature_type': int.from_bytes(tx[8], 'big')
        })

        # Get signature data
        signature_data = rlp.decode(tx[9])
        if struct['signature_type'] == cls.SIGNATURE_SINGLE_TYPE:
            signature_data = {
                'v': int.from_bytes(signature_data[0], 'big'),
                'r': signature_data[1].hex(),
                's': signature_data[2].hex()
            }
        else:
            # Decode signatures
            signatures = []
            for signature in signature_data[1]:
                signatures.append({
                    'v': int.from_bytes(signature[0], 'big'),
                    'r': signature[1].hex(),
                    's': signature[2].hex()
                })

            # Create decoded signature data
            signature_data = {
                'from_mx': MinterHelper.prefix_add(
                    signature_data[0].hex(), PREFIX_ADDR
                ),
                'signatures': signatures
            }
        struct['signature_data'] = signature_data

        # Find out which of tx instance need to create depending on it's type
        data = rlp.decode(tx[5])
        if struct['type'] == MinterDelegateTx.TYPE:
            _class = MinterDelegateTx
        elif struct['type'] == MinterSendCoinTx.TYPE:
            _class = MinterSendCoinTx
        elif struct['type'] == MinterBuyCoinTx.TYPE:
            _class = MinterBuyCoinTx
        elif struct['type'] == MinterCreateCoinTx.TYPE:
            _class = MinterCreateCoinTx
        elif struct['type'] == MinterDeclareCandidacyTx.TYPE:
            _class = MinterDeclareCandidacyTx
        elif struct['type'] == MinterRedeemCheckTx.TYPE:
            _class = MinterRedeemCheckTx
        elif struct['type'] == MinterSellAllCoinTx.TYPE:
            _class = MinterSellAllCoinTx
        elif struct['type'] == MinterSellCoinTx.TYPE:
            _class = MinterSellCoinTx
        elif struct['type'] == MinterSetCandidateOffTx.TYPE:
            _class = MinterSetCandidateOffTx
        elif struct['type'] == MinterSetCandidateOnTx.TYPE:
            _class = MinterSetCandidateOnTx
        elif struct['type'] == MinterUnbondTx.TYPE:
            _class = MinterUnbondTx
        elif struct['type'] == MinterEditCandidateTx.TYPE:
            _class = MinterEditCandidateTx
        elif struct['type'] == MinterMultiSendCoinTx.TYPE:
            _class = MinterMultiSendCoinTx
        else:
            raise Exception('Undefined tx type.')

        # Set tx data
        struct['data'] = _class._data_from_raw(data)

        # Set sender address and raw tx to minter dict
        # ONLY AFTER tx data was set
        struct.update({
            'from_mx': cls.get_sender_address(tx=copy.copy(struct)),
            'signed_tx': raw_tx
        })

        # Prepare **kwargs for creating _class instance.
        # Pass copy of the struct.
        kwargs = _class._structure_to_kwargs(copy.copy(struct))

        return _class(**kwargs)

    @classmethod
    def get_sender_address(cls, tx):
        """
        Get sender address from tx.
        Recover public key from tx and then get address from public key,
        if tx has single signature type, or get decoded sender address,
        if tx has multi signature type.
        Args:
            tx (dict): transaction dict
        Returns:
            Minter address (string)
        """

        # Remember signature data and remove it from tx
        signature_data = tx.pop('signature_data')

        # If there is sender address in signature data (multi signature tx),
        # return it
        if signature_data.get('from_mx'):
            return signature_data['from_mx']

        # Otherwise (single signature tx), recover public key and get
        # address from public key
        # Unhexlify hexdigit dict values to bytes
        tx = MinterHelper.hex2bin_recursive(tx)

        # Encode tx data to RLP
        tx['data'] = rlp.encode(list(tx['data'].values()))

        # Message
        tx_rlp = rlp.encode(list(tx.values()))
        _keccak = MinterHelper.keccak_hash(tx_rlp)

        # Recover public key
        public_key = MinterHelper.prefix_add(
            ECDSA.recover(_keccak, tuple(signature_data.values())),
            PREFIX_PUBKEY
        )

        return MinterWallet.get_address_from_public_key(public_key)

    @classmethod
    def add_signature(cls, signed_tx, private_key):
        """
        Add signature to already signed tx. Method is available only for multisig txs
        Args:
            signed_tx (str): signed tx
            private_key (str): private key
        Returns:
            tx object
        """

        # Create tx instance from raw tx
        tx = cls.from_raw(signed_tx)

        # Check tx to be multi signature type
        if tx.signature_type != cls.SIGNATURE_MULTI_TYPE:
            raise Exception(
                'Signature can be added only to tx with multi signature type'
            )

        # Convert signature data from verbose dict to needed list of
        # signatures
        signature_data = [
            bytes.fromhex(
                MinterHelper.prefix_remove(tx.signature_data['from_mx'])
            ),
            []
        ]
        for item in tx.signature_data['signatures']:
            # Create raw signature (values as integers)
            raw = [
                item['v'],
                int(item['r'], 16),
                int(item['s'], 16)
            ]
            # Append raw signature to total signature data
            signature_data[1].append(raw)

        # Get tx populated structure and keccak hash from this structure
        tx_struct = tx.generate_tx_rlp()

        # Create new signature and update signatures list and signatures
        # attribute of tx
        signature = cls.decode_signature(
            signature=tx.generate_signature(private_key)
        )
        signature_data[1].append(signature)
        tx.signature_data['signatures'].append({
            'v': signature[0],
            'r': format(signature[1], 'x'),
            's': format(signature[2], 'x')
        })

        # Update resulting struct signature data
        tx_struct['signature_data'] = rlp.encode(signature_data)

        # Generate new signed tx and update tx object attribute
        tx.signed_tx = cls.generate_signed_tx(tx_struct)

        return tx

    @classmethod
    def decode_signature(cls, signature):
        """
        Decode hexed signature to raw signature (list of integers)
        Args:
            signature (str): hexed signature
        Returns:
            signature (list[int])
        """
        signature = bytes.fromhex(signature)
        signature = rlp.decode(signature)
        signature = [int.from_bytes(item, 'big') for item in signature]

        return signature

    @classmethod
    def _data_from_raw(cls, raw_data):
        """
        Decoding tx data to tx attributes.
        Decoding depends on tx type, so this method
        must be implemented in each child class.
        """
        raise NotImplementedError


class MinterBuyCoinTx(MinterTx):
    """ Buy coin transaction """

    # Type of transaction
    TYPE = 4

    # Fee units
    COMMISSION = 100

    def __init__(self, coin_to_buy, value_to_buy, coin_to_sell,
                 max_value_to_sell, **kwargs):
        """
        Args:
            coin_to_buy (str): coin name to buy
            value_to_buy (float|int): how much coin to buy (BIP)
            coin_to_sell (str): coin name to sell
            max_value_to_sell (float|int): max amount to sell (BIP)
        """

        super().__init__(**kwargs)

        self.coin_to_buy = coin_to_buy.upper()
        self.value_to_buy = value_to_buy
        self.coin_to_sell = coin_to_sell.upper()
        self.max_value_to_sell = max_value_to_sell

    def _structure_from_instance(self):
        """ Override parent method. """

        struct = super()._structure_from_instance()

        struct.update({
            'type': self.TYPE,
            'data': {
                'coin_to_buy': MinterHelper.encode_coin_name(self.coin_to_buy),
                'value_to_buy': MinterHelper.to_pip(self.value_to_buy),
                'coin_to_sell': MinterHelper.encode_coin_name(self.coin_to_sell),
                'max_value_to_sell': MinterHelper.to_pip(self.max_value_to_sell)
            }
        })

        return struct

    @classmethod
    def _structure_to_kwargs(cls, structure):
        """ Prepare decoded structure data to instance kwargs. """

        kwargs = super()._structure_to_kwargs(structure)

        # Convert data values to verbose.
        # Data will be passed as additional kwarg
        kwargs['data'].update({
            'coin_to_buy': MinterHelper.decode_coin_name(
                kwargs['data']['coin_to_buy']
            ),
            'value_to_buy': MinterHelper.to_bip(
                int.from_bytes(kwargs['data']['value_to_buy'], 'big')
            ),
            'coin_to_sell': MinterHelper.decode_coin_name(
                kwargs['data']['coin_to_sell']
            ),
            'max_value_to_sell': MinterHelper.to_bip(
                int.from_bytes(kwargs['data']['max_value_to_sell'], 'big')
            )
        })

        # Populate data key values as kwargs
        kwargs.update(kwargs['data'])

        return kwargs

    @classmethod
    def _data_from_raw(cls, raw_data):
        """ Parent method implementation """
        return {
            'coin_to_buy': raw_data[0],
            'value_to_buy': raw_data[1],
            'coin_to_sell': raw_data[2],
            'max_value_to_sell': raw_data[3]
        }


class MinterCreateCoinTx(MinterTx):
    """ Create coin transaction """

    # Type of transaction
    TYPE = 5

    # Fee units
    COMMISSION = 1000

    def __init__(self, name, symbol, initial_amount, initial_reserve, crr,
                 max_supply, **kwargs):
        """
        Args:
            name (str): coin name
            symbol (str): coin symbol
            initial_amount (float|int): amount in BIP
            initial_reserve (float|int): reserve in BIP
            crr (int)
        """

        super().__init__(**kwargs)

        self.name = name
        self.symbol = symbol.upper()
        self.initial_amount = initial_amount
        self.initial_reserve = initial_reserve
        self.crr = crr
        self.max_supply = max_supply

    def _structure_from_instance(self):
        """ Override parent method. """

        struct = super()._structure_from_instance()

        struct.update({
            'type': self.TYPE,
            'data': {
                'name': self.name,
                'symbol': MinterHelper.encode_coin_name(self.symbol),
                'initial_amount': MinterHelper.to_pip(self.initial_amount),
                'initial_reserve': MinterHelper.to_pip(self.initial_reserve),
                'crr': '' if self.crr == 0 else self.crr,
                'max_supply': MinterHelper.to_pip(self.max_supply)
            }
        })

        return struct

    @classmethod
    def _structure_to_kwargs(cls, structure):
        """ Prepare decoded structure data to instance kwargs. """

        kwargs = super()._structure_to_kwargs(structure)

        # Convert data values to verbose.
        # Data will be passed as additional kwarg
        kwargs['data'].update({
            'name': kwargs['data']['name'].decode(),
            'symbol': MinterHelper.decode_coin_name(kwargs['data']['symbol']),
            'initial_amount': MinterHelper.to_bip(
                int.from_bytes(kwargs['data']['initial_amount'], 'big')
            ),
            'initial_reserve': MinterHelper.to_bip(
                int.from_bytes(kwargs['data']['initial_reserve'], 'big')
            ),
            'crr': int.from_bytes(kwargs['data']['crr'], 'big'),
            'max_supply': MinterHelper.to_bip(
                int.from_bytes(kwargs['data']['max_supply'], 'big')
            )
        })

        # Populate data key values as kwargs
        kwargs.update(kwargs['data'])

        return kwargs

    @classmethod
    def _data_from_raw(cls, raw_data):
        """ Parent method implementation """
        return {
            'name': raw_data[0],
            'symbol': raw_data[1],
            'initial_amount': raw_data[2],
            'initial_reserve': raw_data[3],
            'crr': raw_data[4],
            'max_supply': raw_data[5]
        }


class MinterDeclareCandidacyTx(MinterTx):
    """ Declare candidacy transaction """

    # Type of transaction
    TYPE = 6

    # Fee units
    COMMISSION = 10000

    def __init__(self, address, pub_key, commission, coin, stake, **kwargs):
        """
        Args:
            address (str): candidate address
            pub_key (str): candidate public key
            commission (int): candidate commission
            coin (str): coin name
            stake (float|int): stake in BIP
        """

        super().__init__(**kwargs)

        self.address = address
        self.pub_key = pub_key
        self.commission = '' if commission == 0 else commission
        self.coin = coin.upper()
        self.stake = stake

    def _structure_from_instance(self):
        """ Override parent method. """

        struct = super()._structure_from_instance()

        struct.update({
            'type': self.TYPE,
            'data': {
                'address': bytes.fromhex(
                    MinterHelper.prefix_remove(self.address)
                ),
                'pub_key': bytes.fromhex(
                    MinterHelper.prefix_remove(self.pub_key)
                ),
                'commission': '' if self.commission == 0 else self.commission,
                'coin': MinterHelper.encode_coin_name(self.coin),
                'stake': MinterHelper.to_pip(self.stake)
            }
        })

        return struct

    @classmethod
    def _structure_to_kwargs(cls, structure):
        """ Prepare decoded structure data to instance kwargs. """

        kwargs = super()._structure_to_kwargs(structure)

        # Convert data values to verbose.
        # Data will be passed as additional kwarg
        kwargs['data'].update({
            'address': MinterHelper.prefix_add(
                kwargs['data']['address'].hex(), PREFIX_ADDR
            ),
            'pub_key': MinterHelper.prefix_add(
                kwargs['data']['pub_key'].hex(), PREFIX_PUBKEY
            ),
            'commission': int.from_bytes(kwargs['data']['commission'], 'big'),
            'coin': MinterHelper.decode_coin_name(kwargs['data']['coin']),
            'stake': MinterHelper.to_bip(
                int.from_bytes(kwargs['data']['stake'], 'big')
            )
        })

        # Populate data key values as kwargs
        kwargs.update(kwargs['data'])

        return kwargs

    @classmethod
    def _data_from_raw(cls, raw_data):
        """ Parent method implementation """
        return {
            'address': raw_data[0],
            'pub_key': raw_data[1],
            'commission': raw_data[2],
            'coin': raw_data[3],
            'stake': raw_data[4]
        }


class MinterDelegateTx(MinterTx):
    """ Delegate transaction """

    # Type of transaction
    TYPE = 7

    # Fee units
    COMMISSION = 200

    def __init__(self, pub_key, coin, stake, **kwargs):
        super().__init__(**kwargs)

        self.pub_key = pub_key
        self.coin = coin.upper()
        self.stake = stake

    def _structure_from_instance(self):
        """ Override parent method to add tx special data. """

        struct = super()._structure_from_instance()

        struct.update({
            'type': self.TYPE,
            'data': {
                'pub_key': bytes.fromhex(
                    MinterHelper.prefix_remove(self.pub_key)
                ),
                'coin': MinterHelper.encode_coin_name(self.coin),
                'stake': MinterHelper.to_pip(self.stake)
            }
        })

        return struct

    @classmethod
    def _structure_to_kwargs(cls, structure):
        """ Prepare decoded structure data to instance kwargs. """
        kwargs = super()._structure_to_kwargs(structure)

        # Convert data values to verbose.
        # Data will be passed as additional kwarg
        kwargs['data'].update({
            'pub_key': MinterHelper.prefix_add(
                kwargs['data']['pub_key'].hex(), PREFIX_PUBKEY
            ),
            'coin': MinterHelper.decode_coin_name(kwargs['data']['coin']),
            'stake': MinterHelper.to_bip(
                int.from_bytes(kwargs['data']['stake'], 'big')
            )
        })

        # Populate data key values as kwargs
        kwargs.update(kwargs['data'])

        return kwargs

    @classmethod
    def _data_from_raw(cls, raw_data):
        """ Parent method implementation """
        return {
            'pub_key': raw_data[0],
            'coin': raw_data[1],
            'stake': raw_data[2]
        }


class MinterRedeemCheckTx(MinterTx):
    """ Redeem check transaction """

    # Type of transaction
    TYPE = 9

    # Fee units
    COMMISSION = 10

    def __init__(self, check, proof, **kwargs):
        """
        Args:
            check (str)
            proof (str)
        """

        super().__init__(**kwargs)

        self.check = check
        self.proof = proof

    def _structure_from_instance(self):
        """ Override parent method to add tx special data. """

        struct = super()._structure_from_instance()

        struct.update({
            'type': self.TYPE,
            'data': {
                'check': bytes.fromhex(MinterHelper.prefix_remove(self.check)),
                'proof': bytes.fromhex(self.proof)
            }
        })

        return struct

    @classmethod
    def _structure_to_kwargs(cls, structure):
        """ Prepare decoded structure data to instance kwargs. """

        kwargs = super()._structure_to_kwargs(structure)

        # Convert data values to verbose.
        # Data will be passed as additional kwarg
        kwargs['data'].update({
            'check': MinterHelper.prefix_add(
                kwargs['data']['check'].hex(), PREFIX_CHECK
            ),
            'proof': kwargs['data']['proof'].hex()
        })

        # Populate data key values as kwargs
        kwargs.update(kwargs['data'])

        return kwargs

    @classmethod
    def _data_from_raw(cls, raw_data):
        """ Parent method implementation """
        return {
            'check': raw_data[0],
            'proof': raw_data[1]
        }


class MinterSellAllCoinTx(MinterTx):
    """ Sell all coin transaction """

    # Type of transaction
    TYPE = 3

    # Fee units
    COMMISSION = 100

    def __init__(self, coin_to_sell, coin_to_buy, min_value_to_buy, **kwargs):
        """
        Args:
            coin_to_sell (str)
            coin_to_buy (str)
            min_value_to_buy (float|int): BIP
        """

        super().__init__(**kwargs)

        self.coin_to_sell = coin_to_sell.upper()
        self.coin_to_buy = coin_to_buy.upper()
        self.min_value_to_buy = min_value_to_buy

    def _structure_from_instance(self):
        """ Override parent method to add tx special data. """

        struct = super()._structure_from_instance()

        struct.update({
            'type': self.TYPE,
            'data': {
                'coin_to_sell': MinterHelper.encode_coin_name(self.coin_to_sell),
                'coin_to_buy': MinterHelper.encode_coin_name(self.coin_to_buy),
                'min_value_to_buy': MinterHelper.to_pip(self.min_value_to_buy)
            }
        })

        return struct

    @classmethod
    def _structure_to_kwargs(cls, structure):
        """ Prepare decoded structure data to instance kwargs. """

        kwargs = super()._structure_to_kwargs(structure)

        # Convert data values to verbose.
        # Data will be passed as additional kwarg
        kwargs['data'].update({
            'coin_to_sell': MinterHelper.decode_coin_name(
                kwargs['data']['coin_to_sell']
            ),
            'coin_to_buy': MinterHelper.decode_coin_name(
                kwargs['data']['coin_to_buy']
            ),
            'min_value_to_buy': MinterHelper.to_bip(
                int.from_bytes(kwargs['data']['min_value_to_buy'], 'big')
            )
        })

        # Populate data key values as kwargs
        kwargs.update(kwargs['data'])

        return kwargs

    @classmethod
    def _data_from_raw(cls, raw_data):
        """ Parent method implementation """
        return {
            'coin_to_sell': raw_data[0],
            'coin_to_buy': raw_data[1],
            'min_value_to_buy': raw_data[2]
        }


class MinterSellCoinTx(MinterTx):
    """ Sell coin transaction """

    # Type of transaction
    TYPE = 2

    # Fee units
    COMMISSION = 100

    def __init__(self, coin_to_sell, value_to_sell, coin_to_buy,
                 min_value_to_buy, **kwargs):
        """
        Args:
            coin_to_sell (str)
            value_to_sell (float|int): BIP
            coin_to_buy (str)
            min_value_to_buy (float|int): BIP
        """

        super().__init__(**kwargs)

        self.coin_to_sell = coin_to_sell.upper()
        self.value_to_sell = value_to_sell
        self.coin_to_buy = coin_to_buy.upper()
        self.min_value_to_buy = min_value_to_buy

    def _structure_from_instance(self):
        """ Override parent method to add tx special data. """

        struct = super()._structure_from_instance()

        struct.update({
            'type': self.TYPE,
            'data': {
                'coin_to_sell': MinterHelper.encode_coin_name(self.coin_to_sell),
                'value_to_sell': MinterHelper.to_pip(self.value_to_sell),
                'coin_to_buy': MinterHelper.encode_coin_name(self.coin_to_buy),
                'min_value_to_buy': MinterHelper.to_pip(self.min_value_to_buy)
            }
        })

        return struct

    @classmethod
    def _structure_to_kwargs(cls, structure):
        """ Prepare decoded structure data to instance kwargs. """

        kwargs = super()._structure_to_kwargs(structure)

        # Convert data values to verbose.
        # Data will be passed as additional kwarg
        kwargs['data'].update({
            'coin_to_sell': MinterHelper.decode_coin_name(
                kwargs['data']['coin_to_sell']
            ),
            'value_to_sell': MinterHelper.to_bip(
                int.from_bytes(kwargs['data']['value_to_sell'], 'big')
            ),
            'coin_to_buy': MinterHelper.decode_coin_name(
                kwargs['data']['coin_to_buy']
            ),
            'min_value_to_buy': MinterHelper.to_bip(
                int.from_bytes(kwargs['data']['min_value_to_buy'], 'big')
            )
        })

        # Populate data key values as kwargs
        kwargs.update(kwargs['data'])

        return kwargs

    @classmethod
    def _data_from_raw(cls, raw_data):
        """ Parent method implementation """
        return {
            'coin_to_sell': raw_data[0],
            'value_to_sell': raw_data[1],
            'coin_to_buy': raw_data[2],
            'min_value_to_buy': raw_data[3]
        }


class MinterSendCoinTx(MinterTx):
    """ Send coin transaction """

    # Type of transaction
    TYPE = 1

    # Fee units
    COMMISSION = 10

    def __init__(self, coin, to, value, **kwargs):
        super().__init__(**kwargs)

        self.coin = coin.upper()
        self.to = to
        self.value = value

    def _structure_from_instance(self):
        """ Override parent method to add tx special data. """

        struct = super()._structure_from_instance()

        struct.update({
            'type': self.TYPE,
            'data': {
                'coin': MinterHelper.encode_coin_name(self.coin),
                'to': bytes.fromhex(MinterHelper.prefix_remove(self.to)),
                'value': MinterHelper.to_pip(self.value)
            }
        })

        return struct

    @classmethod
    def _structure_to_kwargs(cls, structure):
        """ Prepare decoded structure data to instance kwargs. """

        kwargs = super()._structure_to_kwargs(structure)

        # Convert data values to verbose.
        # Data will be passed as additional kwarg
        kwargs['data'].update({
            'coin': MinterHelper.decode_coin_name(kwargs['data']['coin']),
            'to': MinterHelper.prefix_add(
                kwargs['data']['to'].hex(), PREFIX_ADDR
            ),
            'value': MinterHelper.to_bip(
                int.from_bytes(kwargs['data']['value'], 'big')
            )
        })

        # Populate data key values as kwargs
        kwargs.update(kwargs['data'])

        return kwargs

    @classmethod
    def _data_from_raw(cls, raw_data):
        """ Parent method implementation """
        return {
            'coin': raw_data[0],
            'to': raw_data[1],
            'value': raw_data[2]
        }


class MinterMultiSendCoinTx(MinterTx):
    """ Multi send transaction """

    # Type of transaction
    TYPE = 13

    # Fee units
    COMMISSION = 10
    COMMISSION_PER_RECIPIENT = 5

    def __init__(self, txs, **kwargs):
        """
        Args:
            txs (list[dict{coin, to, value}]): list of send coin data
        """
        super().__init__(**kwargs)

        self.txs = txs

    def get_fee(self):
        """
        Override parent method to calc
        multisend-specific fee: (n_txs - 1) * 5 units
        """
        base_fee = super().get_fee()
        recipients_fee = (
            (len(self.txs)-1) * self.COMMISSION_PER_RECIPIENT *
            self.FEE_DEFAULT_MULTIPLIER
        )

        return base_fee + recipients_fee

    def _structure_from_instance(self):
        """ Override parent method to add tx special data. """

        struct = super()._structure_from_instance()

        struct.update({
            'type': self.TYPE,
            'data': {
                'txs': []
            }
        })

        # Populate multi data from each single tx.
        for item in self.txs:
            struct['data']['txs'].append([
                MinterHelper.encode_coin_name(item['coin'].upper()),
                bytes.fromhex(MinterHelper.prefix_remove(item['to'])),
                MinterHelper.to_pip(item['value'])
            ])

        return struct

    @classmethod
    def _structure_to_kwargs(cls, structure):
        """ Prepare decoded structure data to instance kwargs. """

        kwargs = super()._structure_to_kwargs(structure)

        # Convert data values to verbose.
        # Data will be passed as additional kwarg
        for index, item in enumerate(kwargs['data']['txs']):
            kwargs['data']['txs'][index] = {
                'coin': MinterHelper.decode_coin_name(item[0]),
                'to': MinterHelper.prefix_add(item[1].hex(), PREFIX_ADDR),
                'value': MinterHelper.to_pip(int.from_bytes(item[2], 'big'))
            }

        # Populate data key values as kwargs
        kwargs.update(kwargs['data'])

        return kwargs

    @classmethod
    def _data_from_raw(cls, raw_data):
        """ Parent method implementation """
        data = {'txs': []}
        for item in raw_data[0]:
            data['txs'].append([item[0], item[1], item[2]])

        return data


class MinterSetCandidateOffTx(MinterTx):
    """ Set candidate OFF transaction """

    # Type of transaction
    TYPE = 11

    # Fee units
    COMMISSION = 100

    def __init__(self, pub_key, **kwargs):
        """
        Args:
            pub_key (str)
        """

        super().__init__(**kwargs)

        self.pub_key = pub_key

    def _structure_from_instance(self):
        """ Override parent method to add tx special data. """

        struct = super()._structure_from_instance()

        struct.update({
            'type': self.TYPE,
            'data': {
                'pub_key': bytes.fromhex(
                    MinterHelper.prefix_remove(self.pub_key)
                )
            }
        })

        return struct

    @classmethod
    def _structure_to_kwargs(cls, structure):
        """ Prepare decoded structure data to instance kwargs. """

        kwargs = super()._structure_to_kwargs(structure)

        # Convert data values to verbose.
        # Data will be passed as additional kwarg
        kwargs['data'].update({
            'pub_key': MinterHelper.prefix_add(
                kwargs['data']['pub_key'].hex(), PREFIX_PUBKEY
            )
        })

        # Populate data key values as kwargs
        kwargs.update(kwargs['data'])

        return kwargs

    @classmethod
    def _data_from_raw(cls, raw_data):
        """ Parent method implementation """
        return {
            'pub_key': raw_data[0]
        }


class MinterSetCandidateOnTx(MinterTx):
    """ Set candidate ON transaction """

    # Type of transaction
    TYPE = 10

    # Fee units
    COMMISSION = 100

    def __init__(self, pub_key, **kwargs):
        """
        Args:
            pub_key (str)
        """

        super().__init__(**kwargs)

        self.pub_key = pub_key

    def _structure_from_instance(self):
        """ Override parent method to add tx special data. """

        struct = super()._structure_from_instance()

        struct.update({
            'type': self.TYPE,
            'data': {
                'pub_key': bytes.fromhex(
                    MinterHelper.prefix_remove(self.pub_key)
                )
            }
        })

        return struct

    @classmethod
    def _structure_to_kwargs(cls, structure):
        """ Prepare decoded structure data to instance kwargs. """

        kwargs = super()._structure_to_kwargs(structure)

        # Convert data values to verbose.
        # Data will be passed as additional kwarg
        kwargs['data'].update({
            'pub_key': MinterHelper.prefix_add(
                kwargs['data']['pub_key'].hex(), PREFIX_PUBKEY
            )
        })

        # Populate data key values as kwargs
        kwargs.update(kwargs['data'])

        return kwargs

    @classmethod
    def _data_from_raw(cls, raw_data):
        """ Parent method implementation """
        return {
            'pub_key': raw_data[0]
        }


class MinterUnbondTx(MinterTx):
    """ Unbond transaction """

    # Type of transaction
    TYPE = 8

    # Fee units
    COMMISSION = 100

    def __init__(self, pub_key, coin, value, **kwargs):
        """
        Args:
            pub_key (str)
            coin (str)
            value (float|int): BIP
        """

        super().__init__(**kwargs)

        self.pub_key = pub_key
        self.coin = coin.upper()
        self.value = value

    def _structure_from_instance(self):
        """ Override parent method to add tx special data. """

        struct = super()._structure_from_instance()

        struct.update({
            'type': self.TYPE,
            'data': {
                'pub_key': bytes.fromhex(
                    MinterHelper.prefix_remove(self.pub_key)
                ),
                'coin': MinterHelper.encode_coin_name(self.coin),
                'value': MinterHelper.to_pip(self.value)
            }
        })

        return struct

    @classmethod
    def _structure_to_kwargs(cls, structure):
        """ Prepare decoded structure data to instance kwargs. """

        kwargs = super()._structure_to_kwargs(structure)

        # Convert data values to verbose.
        # Data will be passed as additional kwarg
        kwargs['data'].update({
            'pub_key': MinterHelper.prefix_add(
                kwargs['data']['pub_key'].hex(), PREFIX_PUBKEY
            ),
            'coin': MinterHelper.decode_coin_name(kwargs['data']['coin']),
            'value': MinterHelper.to_bip(
                int.from_bytes(kwargs['data']['value'], 'big')
            )
        })

        # Populate data key values as kwargs
        kwargs.update(kwargs['data'])

        return kwargs

    @classmethod
    def _data_from_raw(cls, raw_data):
        """ Parent method implementation """
        return {
            'pub_key': raw_data[0],
            'coin': raw_data[1],
            'value': raw_data[2]
        }


class MinterEditCandidateTx(MinterTx):
    """ Edit candidate transaction """

    # Type of transaction
    TYPE = 14

    # Fee units
    COMMISSION = 10000

    def __init__(self, pub_key, reward_address, owner_address, **kwargs):
        """
        Args:
            pub_key (str): candidate public key
            reward_address (str)
            owner_address (str)
        """

        super().__init__(**kwargs)

        self.pub_key = pub_key
        self.reward_address = reward_address
        self.owner_address = owner_address

    def _structure_from_instance(self):
        """ Override parent method to add tx special data. """

        struct = super()._structure_from_instance()

        struct.update({
            'type': self.TYPE,
            'data': {
                'pub_key': bytes.fromhex(
                    MinterHelper.prefix_remove(self.pub_key)
                ),
                'reward_address': bytes.fromhex(
                    MinterHelper.prefix_remove(self.reward_address)
                ),
                'owner_address': bytes.fromhex(
                    MinterHelper.prefix_remove(self.owner_address)
                )
            }
        })

        return struct

    @classmethod
    def _structure_to_kwargs(cls, structure):
        """ Prepare decoded structure data to instance kwargs. """

        kwargs = super()._structure_to_kwargs(structure)

        # Convert data values to verbose.
        # Data will be passed as additional kwarg
        kwargs['data'].update({
            'pub_key': MinterHelper.prefix_add(
                kwargs['data']['pub_key'].hex(), PREFIX_PUBKEY
            ),
            'reward_address': MinterHelper.prefix_add(
                kwargs['data']['reward_address'].hex(), PREFIX_ADDR
            ),
            'owner_address': MinterHelper.prefix_add(
                kwargs['data']['owner_address'].hex(), PREFIX_ADDR
            )
        })

        # Populate data key values as kwargs
        kwargs.update(kwargs['data'])

        return kwargs

    @classmethod
    def _data_from_raw(cls, raw_data):
        """ Parent method implementation """
        return {
            'pub_key': raw_data[0],
            'reward_address': raw_data[1],
            'owner_address': raw_data[2]
        }



import setuptools
import codecs

with codecs.open("README.md", "r", 'utf_8_sig') as fh:
    long_description = fh.read()

setuptools.setup(
    name="minter-sdk",
    version="1.0.29",
    author="U-node Team",
    author_email="rymka1989@gmail.com",
    description=u"Python SDK for Minter Network",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/U-node/minter-sdk",
    packages=setuptools.find_packages(include=['mintersdk']),
    include_package_data=True,
    classifiers=(
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ),
    install_requires=[
        'rlp',
        'sslcrypto',
        'mnemonic',
        'pysha3',
        'requests',
        'pyqrcode',
        'deprecated'
    ]
)
